#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# External Libraries
import numpy as np
import tensorflow as tf
# Internal Libraries
from aucmedi.xai.methods.xai_base import XAImethod_Base

#-----------------------------------------------------#
#                 Integrated Gradients                #
#-----------------------------------------------------#
class IntegratedGradients(XAImethod_Base):
    """ XAI Method for Integrated Gradients.

    Normally, this class is used internally in the [aucmedi.xai.decoder.xai_decoder][] in the AUCMEDI XAI module.

    ??? abstract "Reference - Implementation"
        Author: Aakash Kumar Nain <br>
        GitHub Profile: [https://github.com/AakashKumarNain](https://github.com/AakashKumarNain) <br>
        Date: Jun 02, 2020 <br>
        [https://keras.io/examples/vision/integrated_gradients](https://keras.io/examples/vision/integrated_gradients) <br>

    ??? abstract "Reference - Publication"
        Mukund Sundararajan, Ankur Taly, Qiqi Yan. 04 Mar 2017.
        Axiomatic Attribution for Deep Networks.
        <br>
        [https://arxiv.org/abs/1703.01365](https://arxiv.org/abs/1703.01365)

    This class provides functionality for running the compute_heatmap function,
    which computes an Integrated Gradients Map for an image with a model.
    """
    def __init__(self, model, layerName=None, num_steps=50):
        """ Initialization function for creating a Integrated Gradients Map as XAI Method object.

        Args:
            model (keras.model):            Keras model object.
            layerName (str):                Not required in Integrated Gradients Maps, but defined by Abstract Base Class.
            num_steps (int):                Number of iterations for interpolation.
        """
        # Cache class parameters
        self.model = model
        self.num_steps = num_steps

    #---------------------------------------------#
    #             Heatmap Computation             #
    #---------------------------------------------#
    def compute_heatmap(self, image, class_index, eps=1e-8):
        """ Core function for computing the Integrated Gradients Map for a provided image and for specific classification outcome.

        ???+ attention
            Be aware that the image has to be provided in batch format.

        Args:
            image (numpy.ndarray):              Image matrix encoded as NumPy Array (provided as one-element batch).
            class_index (int):                  Classification index for which the heatmap should be computed.
            eps (float):                        Epsilon for rounding.

        The returned heatmap is encoded within a range of [0,1]

        ???+ attention
            The shape of the returned heatmap is 2D -> batch and channel axis will be removed.

        Returns:
            heatmap (numpy.ndarray):            Computed Integrated Gradients Map for provided image.
        """
        # Perform interpolation
        baseline = np.zeros(image.shape).astype(np.float32)
        interpolated_imgs = []
        for step in range(0, self.num_steps + 1):
            cii = baseline + (step / self.num_steps) * (image - baseline)
            interpolated_imgs.append(cii)
        interpolated_imgs = np.array(interpolated_imgs).astype(np.float32)

        # Get the gradients for each interpolated image
        grads = []
        for int_img in interpolated_imgs:
            # Compute gradient
            with tf.GradientTape() as tape:
                inputs = tf.cast(int_img, tf.float32)
                tape.watch(inputs)
                preds = self.model(inputs)
                loss = preds[:, class_index]
            gradient = tape.gradient(loss, inputs)
            # Add to gradient list
            grads.append(gradient[0])
        grads = tf.convert_to_tensor(grads, dtype=tf.float32)

        # Approximate the integral using the trapezoidal rule
        grads = (grads[:-1] + grads[1:]) / 2.0
        avg_grads = tf.reduce_mean(grads, axis=0)
        # Calculate integrated gradients
        integrated_grads = (image - baseline) * avg_grads
        # Obtain maximum gradient
        integrated_grads = tf.reduce_max(integrated_grads, axis=-1)

        # Convert to NumPy & Remove batch axis
        heatmap = integrated_grads.numpy()[0,:,:]
        # Intensity normalization to [0,1]
        numer = heatmap - np.min(heatmap)
        denom = (heatmap.max() - heatmap.min()) + eps
        heatmap = numer / denom

        # Return the resulting heatmap
        return heatmap



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# External libraries
from tensorflow.keras.models import Model
from tensorflow.keras import Input, layers

#-----------------------------------------------------#
#                 Classification Head                 #
#-----------------------------------------------------#
class Classifier:
    """ A powerful interface for all types of image classifications.

    This class will be created automatically inside the [NeuralNetwork][aucmedi.neural_network.model.NeuralNetwork] class.

    !!! info "Supported Features"
        - Binary classification
        - Multi-class classification
        - Multi-label classification
        - 2D/3D data
        - Metadata encoded as NumPy arrays (int or float)

    This class provides functionality for building a classification head for an
    [Architecture][aucmedi.neural_network.architectures]
    ([tensorflow.keras model](https://www.tensorflow.org/api_docs/python/tf/keras/Model)).
    A initialized classifier interface is passed to an architecture class.
    The `build()` function of the classification head is called in the `create_model()`
    function of the architecture.

    !!! info "Structure of the AUCMEDI Classification Head"
        | Layer                         | Description                                                      |
        | ----------------------------- | ---------------------------------------------------------------- |
        | GlobalAveragePooling          | Pooling from Architecture Output to a single spatial dimensions. |
        | Dense(units=512)              | Optional dense & dropout layer if `fcl_dropout=True`.            |
        | Dropout(0.3)                  | Optional dense & dropout layer if `fcl_dropout=True`.            |
        | Concatenate()                 | Optional appending of metadata to classification head.           |
        | Dense(units=512)              | Optional dense & dropout layer if metadata is present.           |
        | Dropout(0.3)                  | Optional dense & dropout layer if metadata is present.           |
        | Dense(units=256)              | Optional dense & dropout layer if metadata is present.           |
        | Dropout(0.3)                  | Optional dense & dropout layer if metadata is present.           |
        | Dense(units=n_labels)         | Dense layer to the number of labels (classes).                   |
        | Activation(activation_output) | Activation function corresponding to classification type.        |

    ???+ note "Classification Types"
        | Type                       | Activation Function                                             |
        | -------------------------- | --------------------------------------------------------------- |
        | Binary classification      | `activation_output="softmax"`: Only a single class is correct.  |
        | Multi-class classification | `activation_output="softmax"`: Only a single class is correct.  |
        | Multi-label classification | `activation_output="sigmoid"`: Multiple classes can be correct. |

        For more information on multi-class vs multi-label, check out this blog post from Rachel Draelos: <br>
        [https://glassboxmedicine.com/2019/05/26/classification-sigmoid-vs-softmax/](https://glassboxmedicine.com/2019/05/26/classification-sigmoid-vs-softmax/)

    The recommended way is to pass all required variables to the [NeuralNetwork][aucmedi.neural_network.model.NeuralNetwork]
    which automatically creates the Classifier and passes it to the Architecture.

    ???+ example
        ```python
        # Recommended way (automatic creation in NeuralNetwork)
        model = NeuralNetwork(n_labels=20, channels=3,
                              input_shape=(32, 32), activation_output="sigmoid",
                              fcl_dropout=False)

        # Manual way
        from aucmedi.neural_network.architectures import Classifier
        from aucmedi.neural_network.architectures.image import Vanilla

        classification_head = Classifier(n_labels=20, fcl_dropout=False,
                                         activation_output="sigmoid")
        arch = Vanilla(classification_head, channels=3,
                       input_shape=(32, 32))
        ```

    ??? example "Example: How to integrate metadata in AUCMEDI?"
        ```python
        from aucmedi import *
        import numpy as np

        my_metadata = np.random.rand(len(samples), 10)

        my_model = NeuralNetwork(n_labels=8, channels=3, architecture="2D.DenseNet121",
                                  meta_variables=10)

        my_dg = DataGenerator(samples, "images_dir/",
                              labels=None, metadata=my_metadata,
                              resize=my_model.meta_input,                  # (224,224)
                              standardize_mode=my_model.meta_standardize)  # "torch"
        ```
    """
    #---------------------------------------------#
    #                Initialization               #
    #---------------------------------------------#
    def __init__(self, n_labels, activation_output="softmax",
                 meta_variables=None, fcl_dropout=True):
        """ Initialization function for creating a Classifier object.

        The fully connected layer and dropout option (`fcl_dropout`) utilizes a 512 unit Dense layer with 30% Dropout.

        Modi for activation_output: Check out [TensorFlow.Keras doc on activation functions](https://www.tensorflow.org/api_docs/python/tf/keras/activations).

        Args:
            n_labels (int):                 Number of classes/labels (important for the last layer of classification head).
            activation_output (str):        Activation function which is used in the last classification layer.
            meta_variables (int):           Number of metadata variables, which should be included in the classification head.
                                            If `None`is provided, no metadata integration block will be added to the classification head.
            fcl_dropout (bool):             Option whether to utilize a Dense & Dropout layer before the last classification layer.
        """
        self.n_labels = n_labels
        self.activation_output = activation_output
        self.meta_variables = meta_variables
        self.fcl_dropout = fcl_dropout

    #---------------------------------------------#
    #                Create Model                 #
    #---------------------------------------------#
    def build(self, model_input, model_output):
        """ Internal function which appends the classification head.

        This function will be called from inside an [Architecture][aucmedi.neural_network.architectures] `create_model()` function
        and must return a functional Keras model.
        The `build()` function will append a classification head to the provided Keras model.

        Args:
            model_input (tf.keras layer):       Input layer of the model.
            model_output (tf.keras layer):      Output layer of the model.

        Returns:
            model (tf.keras model):             A functional Keras model.
        """
        # Apply GlobalAveragePooling to obtain a single spatial dimensions
        if len(model_output.shape) == 4:            # for 2D architectures
            model_head = layers.GlobalAveragePooling2D(name="avg_pool")(model_output)
        elif len(model_output.shape) == 5:          # for 3D architectures
            model_head = layers.GlobalAveragePooling3D(name="avg_pool")(model_output)
        # if not model output shape 4 or 5 -> it is already GlobalAveragePooled to 2 dim
        else : model_head = model_output

        # Apply optional dense & dropout layer
        if self.fcl_dropout:
            model_head = layers.Dense(units=512)(model_head)
            model_head = layers.Dropout(0.3)(model_head)

        # Apply metadata integration block
        if self.meta_variables is not None:
            # Define metadata input
            model_meta = Input(shape=(self.meta_variables,))

            # Integrate metadata into classification had
            model_head = layers.concatenate([model_head, model_meta])

            # Apply additional densely-connected NN layers
            model_head = layers.Dense(units=512, activation="relu")(model_head)
            model_head = layers.Dropout(0.3)(model_head)
            model_head = layers.Dense(units=256, activation="relu")(model_head)
            model_head = layers.Dropout(0.3)(model_head)

        # Apply classifier
        model_head = layers.Dense(self.n_labels, name="preds")(model_head)
        # Apply activation output according to classification type
        model_head = layers.Activation(self.activation_output, name="probs")(model_head)

        # Obtain input layer
        if self.meta_variables is not None:
            input_layer = [model_input, model_meta]
        else : input_layer = model_input

        # Create tf.keras model
        model = Model(inputs=input_layer, outputs=model_head)

        # Return ready-to-use classifier model
        return model



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# External libraries
import os
import numpy as np

#-----------------------------------------------------#
#             Numpy Loader for AUCMEDI IO             #
#-----------------------------------------------------#
def numpy_loader(sample, path_imagedir, image_format=None, grayscale=False,
                 two_dim=True, **kwargs):
    """ NumPy Loader for image loading within the AUCMEDI pipeline.

    The NumPy Loader is an IO_loader function, which have to be passed to the
    [DataGenerator][aucmedi.data_processing.data_generator.DataGenerator].

    The NumPy load function `np.load(path_img, allow_pickle=True)` is used.

    ???+ example
        ```python
        # Import required libraries
        from aucmedi import *
        from aucmedi.data_processing.io_loader import numpy_loader

        # Initialize input data reader
        ds = input_interface(interface="csv",
                             path_imagedir="dataset/npy_files/",
                             path_data="dataset/annotations.csv",
                             ohe=False, col_sample="ID", col_class="diagnosis")
        (samples, class_ohe, nclasses, class_names, image_format) = ds

        # Initialize DataGenerator with numpy_loader
        data_gen = DataGenerator(samples, "dataset/npy_files/", labels=class_ohe,
                                 image_format=image_format, resize=None,
                                 grayscale=True, two_dim=False,
                                 loader=numpy_loader)
        ```

    Args:
        sample (str):               Sample name/index of an image.
        path_imagedir (str):        Path to the directory containing the images.
        image_format (str):         Image format to add at the end of the sample index for image loading.
        grayscale (bool):           Boolean, whether images are grayscale or RGB.
        two_dim (bool):             Boolean, whether image is 2D or 3D.
        **kwargs (dict):            Additional parameters for the sample loader.
    """
    # Get image path
    if image_format : img_file = sample + "." + image_format
    else : img_file = sample
    path_img = os.path.join(path_imagedir, img_file)
    # Load image via the NumPy package
    img = np.load(path_img, allow_pickle=True)
    # Verify image shape for grayscale & 2D
    if grayscale and two_dim:
        # Add channel axis and return image
        if len(img.shape) == 2:
            return np.reshape(img, img.shape + (1,))
        # Just return image
        elif len(img.shape) == 3 and img.shape[-1] == 1:
            return img
        # Throw Exception
        else:
            raise ValueError("Parameter 2D & Grayscale: Expected either 2D " + \
                             "without channel axis or 3D with single channel" + \
                             " axis, but got:", img.shape, len(img.shape))
    # Verify image shape for grayscale & 3D
    elif grayscale and not two_dim:
        # Add channel axis and return image
        if len(img.shape) == 3:
            return np.reshape(img, img.shape + (1,))
        # Just return image
        elif len(img.shape) == 4 and img.shape[-1] == 1:
            return img
        # Throw Exception
        else:
            raise ValueError("Parameter 3D & Grayscale: Expected either 3D " + \
                             "without channel axis or 4D with single channel" + \
                             " axis, but got:", img.shape, len(img.shape))
    # Verify image shape for rgb & 2D
    elif not grayscale and two_dim:
        # Just return image
        if len(img.shape) == 3 and img.shape[-1] == 3:
            return img
        # Throw Exception
        else:
            raise ValueError("Parameter 2D & RGB: Expected 3D array " + \
                             "including a single channel axis, but got:",
                             img.shape, len(img.shape))
    # Verify image shape for rgb & 3D
    elif not grayscale and not two_dim:
        # Just return image
        if len(img.shape) == 4 and img.shape[-1] == 3:
            return img
        # Throw Exception
        else:
            raise ValueError("Parameter 3D & RGB: Expected 4D array " + \
                             "including a single channel axis, but got:",
                             img.shape, len(img.shape))



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# External libraries
import os
import tempfile
from tensorflow.keras.callbacks import ModelCheckpoint, CSVLogger
from pathos.helpers import mp   # instead of 'import multiprocessing as mp'
import numpy as np
import shutil
# Internal libraries
from aucmedi import DataGenerator, NeuralNetwork
from aucmedi.sampling import sampling_split
from aucmedi.ensemble.aggregate import aggregate_dict
from aucmedi.ensemble.metalearner import metalearner_dict
from aucmedi.ensemble.metalearner.ml_base import Metalearner_Base
from aucmedi.ensemble.aggregate.agg_base import Aggregate_Base

#-----------------------------------------------------#
#             Ensemble Learning: Stacking             #
#-----------------------------------------------------#
class Stacking:
    """ A Stacking class providing functionality for metalearner based ensemble learning.

    In contrast to single algorithm approaches, the ensemble of different deep convolutional neural network architectures
    (also called heterogeneous ensemble learning) showed strong benefits for overall performance in several studies.
    The idea of the Stacking technique is to utilize diverse and independent models by stacking another machine learning
    algorithm on top of these predictions.

    In AUCMEDI, a percentage split is applied on the dataset into the subsets: train, validation and ensemble.

    On the train and validation subsets, one or multiple predefined [NeuralNetwork][aucmedi.neural_network.model]
    models are trained, whereas on the ensemble subset a metalearner is fitted to merge
    [NeuralNetwork][aucmedi.neural_network.model] model predictions into a single one.

    ???+ example
        ```python
        # Initialize some NeuralNetwork models
        model_a = NeuralNetwork(n_labels=4, channels=3, architecture="2D.ResNet50")
        model_b = NeuralNetwork(n_labels=4, channels=3, architecture="2D.MobileNetV2")
        model_c = NeuralNetwork(n_labels=4, channels=3, architecture="2D.EfficientNetB1")

        # Initialize Stacking object
        el = Stacking(model_list=[model_a, model_b, model_c],
                      metalearner="logistic_regression")

        # Initialize training DataGenerator for complete training data
        datagen = DataGenerator(samples_train, "images_dir/",
                                labels=train_labels_ohe, batch_size=3,
                                resize=None, standardize_mode=None)
        # Train neural network and metalearner models
        el.train(datagen, epochs=100)

        # Initialize testing DataGenerator for testing data
        test_gen = DataGenerator(samples_test, "images_dir/",
                                 resize=None, standardize_mode=None)
        # Run Inference
        preds = el.predict(test_gen)
        ```

    !!! warning "Training Time Increase"
        Stacking sequentially performs fitting processes for multiple models, which will drastically increase training time.

    ??? warning "DataGenerator re-initialization"
        The passed DataGenerator for the train() and predict() function of the Stacking class will be re-initialized!

        This can result in redundant image preparation if `prepare_images=True`.

        Furthermore, the parameters `resize` and `standardize_mode` are automatically re-initialized with
        NeuralNetwork model specific values (`model.meta_standardize` for `standardize_mode` and
        `model.meta_input` for `input_shape`).

        If desired (but not recommended!), it is possible to modify the meta variables of the NeuralNetwork model as follows:
        ```python
        # For input_shape
        model_a = NeuralNetwork(n_labels=4, channels=3, architecture="2D.ResNet50",
                                input_shape=(64,64))
        # For standardize_mode
        model_b = NeuralNetwork(n_labels=4, channels=3, architecture="2D.MobileNetV2")
        model_b.meta_standardize = "torch"
        ```

    ??? warning "NeuralNetwork re-initialization"
        The passed NeuralNetwork for the train() and predict() function of the Composite class will be re-initialized!

        Attention: Metrics are not passed to the processes due to pickling issues.

    ??? info "Technical Details"
        For the training and inference process, each model will create an individual process via the Python multiprocessing package.

        This is crucial as TensorFlow does not fully support the VRAM memory garbage collection in GPUs,
        which is why more and more redundant data pile up with an increasing number of models.

        Via separate processes, it is possible to clean up the TensorFlow environment and rebuild it again for the next model.

    ??? reference "Reference for Ensemble Learning Techniques"
        Dominik Müller, Iñaki Soto-Rey and Frank Kramer. (2022).
        An Analysis on Ensemble Learning optimized Medical Image Classification with Deep Convolutional Neural Networks.
        arXiv e-print: [https://arxiv.org/abs/2201.11440](https://arxiv.org/abs/2201.11440)
    """
    def __init__(self, model_list, metalearner="logistic_regression",
                 sampling=[0.7, 0.1, 0.2]):
        """ Initialization function for creating a Stacking object.

        Args:
            model_list (list of NeuralNetwork):        List of instances of AUCMEDI neural network class.
            metalearner (str, Metalearner or Aggregate):Metalearner class instance / a string for an AUCMEDI Metalearner,
                                                        or Aggregate function / a string for an AUCMEDI Aggregate function.
            sampling (list of float):                   List of percentage values with split sizes. Should be 3x percentage values
                                                        for heterogenous metalearner and 2x percentage values for homogeneous
                                                        Aggregate functions (must sum up to 1.0).
        """
        # Cache class variables
        self.model_list = model_list
        self.metalearner = metalearner
        self.sampling = sampling
        self.sampling_seed = 0
        self.cache_dir = None

        # Initialize Metalearner
        if isinstance(metalearner, str) and metalearner in metalearner_dict:
            self.ml_model = metalearner_dict[metalearner]()
        elif isinstance(metalearner, str) and metalearner in aggregate_dict:
            self.ml_model = aggregate_dict[metalearner]()
        elif isinstance(metalearner, Metalearner_Base) or \
             isinstance(metalearner, Aggregate_Base):
            self.ml_model = metalearner
        else : raise TypeError("Unknown type of Metalearner (neither known " + \
                               "ensembler nor Aggregate or Metalearner class)!")

        # Set multiprocessing method to spawn
        mp.set_start_method("spawn", force=True)

    def train(self, training_generator, epochs=20, iterations=None,
              callbacks=[], class_weights=None, transfer_learning=False,
              metalearner_fitting=True):
        """ Training function for fitting the provided Stacking models.

        The training data will be sampled according to a percentage split in which
        [DataGenerators][aucmedi.data_processing.data_generator.DataGenerator] for model training
        and validation as well as one for the metalearner training will be automatically created.

        It is also possible to pass custom Callback classes in order to obtain more information.

        For more information on the fitting process, check out [NeuralNetwork.train()][aucmedi.neural_network.model.NeuralNetwork.train].

        Args:
            training_generator (DataGenerator):     A data generator which will be used for training (will be split according
                                                    to percentage split sampling).
            epochs (int):                           Number of epochs. A single epoch is defined as one iteration through
                                                    the complete data set.
            iterations (int):                       Number of iterations (batches) in a single epoch.
            callbacks (list of Callback classes):   A list of Callback classes for custom evaluation.
            class_weights (dictionary or list):     A list or dictionary of float values to handle class unbalance.
            transfer_learning (bool):               Option whether a transfer learning training should be performed.
            metalearner_fitting (bool):             Option whether the Metalearner fitting process should be included in the
                                                    Stacking training process. The `train_metalearner()` function can also be
                                                    run manually (or repeatedly).
        Returns:
            history (dict):                   A history dictionary from a Keras history object which contains several logs.
        """
        temp_dg = training_generator    # Template DataGenerator variable for faster access
        history_stacking = {}           # Final history dictionary

        # Create temporary model directory
        self.cache_dir = tempfile.TemporaryDirectory(prefix="aucmedi.tmp.",
                                                     suffix=".stacking")

        # Obtain training data
        x = training_generator.samples
        y = training_generator.labels
        m = training_generator.metadata

        # Apply percentage split sampling
        ps_sampling = sampling_split(x, y, m, sampling=self.sampling,
                                     stratified=True, iterative=True,
                                     seed=self.sampling_seed)

        # Pack data according to sampling
        if len(ps_sampling[0]) == 3:
            data_train = ps_sampling[0]
            data_val = ps_sampling[1]
        else:
            data_train = (*ps_sampling[0], None)
            data_val = (*ps_sampling[1], None)

        # Sequentially iterate over model list
        for i in range(len(self.model_list)):
            # Create model specific callback list
            callbacks_model = callbacks.copy()
            # Extend Callback list
            path_model = os.path.join(self.cache_dir.name,
                                      "nn_" + str(i) + ".model.keras")
            cb_mc = ModelCheckpoint(path_model,
                                    monitor="val_loss", verbose=1,
                                    save_best_only=True, mode="min")
            cb_cl = CSVLogger(os.path.join(self.cache_dir.name,
                                                 "nn_" + str(i) + \
                                                 ".logs.csv"),
                              separator=',', append=True)
            callbacks_model.extend([cb_mc, cb_cl])

            # Gather NeuralNetwork parameters
            model_paras = {
                "n_labels": self.model_list[i].n_labels,
                "channels": self.model_list[i].channels,
                "input_shape": self.model_list[i].input_shape,
                "architecture": self.model_list[i].architecture,
                "pretrained_weights": self.model_list[i].pretrained_weights,
                "loss": self.model_list[i].loss,
                "metrics": None,
                "activation_output": self.model_list[i].activation_output,
                "fcl_dropout": self.model_list[i].fcl_dropout,
                "meta_variables": self.model_list[i].meta_variables,
                "learning_rate": self.model_list[i].learning_rate,
            }

            # Gather DataGenerator parameters
            datagen_paras = {"path_imagedir": temp_dg.path_imagedir,
                             "batch_size": temp_dg.batch_size,
                             "data_aug": temp_dg.data_aug,
                             "seed": temp_dg.seed,
                             "subfunctions": temp_dg.subfunctions,
                             "shuffle": temp_dg.shuffle,
                             "standardize_mode": self.model_list[i].meta_standardize,
                             "resize": self.model_list[i].meta_input,
                             "grayscale": temp_dg.grayscale,
                             "prepare_images": temp_dg.prepare_images,
                             "sample_weights": temp_dg.sample_weights,
                             "image_format": temp_dg.image_format,
                             "loader": temp_dg.sample_loader,
                             "workers": temp_dg.workers,
                             "kwargs": temp_dg.kwargs
            }

            # Gather training parameters
            parameters_training = {"epochs": epochs,
                                "iterations": iterations,
                                "callbacks": callbacks_model,
                                "class_weights": class_weights,
                                "transfer_learning": transfer_learning
            }

            # Start training process
            process_queue = mp.Queue()
            process_train = mp.Process(target=__training_process__,
                                       args=(process_queue,
                                             model_paras,
                                             data_train,
                                             data_val,
                                             datagen_paras,
                                             parameters_training))
            process_train.start()
            process_train.join()
            nn_history = process_queue.get()
            # Combine logged history objects
            hnn = {"nn_" + str(i) + "." + k: v for k, v in nn_history.items()}
            history_stacking = {**history_stacking, **hnn}

        # Perform metalearner model training
        self.train_metalearner(temp_dg)

        # Return Stacking history object
        return history_stacking

    def train_metalearner(self, training_generator):
        """ Training function for fitting the Metalearner model.

        Function will be called automatically in the `train()` function if
        the parameter `metalearner_fitting` is true.

        However, this function can also be called multiple times for training
        different Metalearner types without the need of time-extensive
        re-training of the [NeuralNetwork][aucmedi.neural_network.model] models.

        Args:
            training_generator (DataGenerator):     A data generator which will be used for training (will be split according
                                                    to percentage split sampling).
        """
        # Skipping metalearner training if aggregate function
        if isinstance(self.ml_model, Aggregate_Base) : return

        temp_dg = training_generator    # Template DataGenerator variable for faster access
        preds_ensemble = []

        # Obtain training data
        x = training_generator.samples
        y = training_generator.labels
        m = training_generator.metadata

        # Apply percentage split sampling
        ps_sampling = sampling_split(x, y, m, sampling=self.sampling,
                                     stratified=True, iterative=True,
                                     seed=self.sampling_seed)

        # Pack data according to sampling
        if len(ps_sampling[0]) == 3 : data_ensemble = ps_sampling[2]
        else : data_ensemble = (*ps_sampling[2], None)

        # Identify path to model directory
        if isinstance(self.cache_dir, tempfile.TemporaryDirectory):
            path_model_dir = self.cache_dir.name
        else : path_model_dir = self.cache_dir

        # Sequentially iterate over model list
        for i in range(len(self.model_list)):
            #  Load current model
            path_model = os.path.join(path_model_dir,
                                      "nn_" + str(i) + ".model.keras")

            # Gather NeuralNetwork parameters
            model_paras = {
                "n_labels": self.model_list[i].n_labels,
                "channels": self.model_list[i].channels,
                "input_shape": self.model_list[i].input_shape,
                "architecture": self.model_list[i].architecture,
                "pretrained_weights": self.model_list[i].pretrained_weights,
                "loss": self.model_list[i].loss,
                "metrics": None,
                "activation_output": self.model_list[i].activation_output,
                "fcl_dropout": self.model_list[i].fcl_dropout,
                "meta_variables": self.model_list[i].meta_variables,
                "learning_rate": self.model_list[i].learning_rate,
            }

            # Gather DataGenerator parameters
            datagen_paras = {"path_imagedir": temp_dg.path_imagedir,
                             "batch_size": temp_dg.batch_size,
                             "data_aug": temp_dg.data_aug,
                             "seed": temp_dg.seed,
                             "subfunctions": temp_dg.subfunctions,
                             "shuffle": temp_dg.shuffle,
                             "standardize_mode": self.model_list[i].meta_standardize,
                             "resize": self.model_list[i].meta_input,
                             "grayscale": temp_dg.grayscale,
                             "prepare_images": temp_dg.prepare_images,
                             "sample_weights": temp_dg.sample_weights,
                             "image_format": temp_dg.image_format,
                             "loader": temp_dg.sample_loader,
                             "workers": temp_dg.workers,
                             "kwargs": temp_dg.kwargs
            }

            # Start inference process for model i
            process_queue = mp.Queue()
            process_pred = mp.Process(target=__prediction_process__,
                                      args=(process_queue,
                                            model_paras,
                                            path_model,
                                            data_ensemble,
                                            datagen_paras))
            process_pred.start()
            process_pred.join()
            preds = process_queue.get()

            # Append preds to ensemble
            preds_ensemble.append(preds)

        # Preprocess prediction ensemble
        preds_ensemble = np.array(preds_ensemble)
        preds_ensemble = np.swapaxes(preds_ensemble, 0, 1)
        s, m, c = preds_ensemble.shape
        x_stack = np.reshape(preds_ensemble, (s, m*c))

        # Start training of stacked metalearner
        if isinstance(self.ml_model, Metalearner_Base):
            (_, y_stack, _) = data_ensemble
            self.ml_model.train(x_stack, y_stack)
            # Store metalearner model to disk
            path_metalearner = os.path.join(path_model_dir,
                                            "metalearner.model.pickle")
            self.ml_model.dump(path_metalearner)

    def predict(self, prediction_generator, return_ensemble=False):
        """ Prediction function for Stacking.

        The fitted models and selected Metalearner will predict classifications for the provided
        [DataGenerator][aucmedi.data_processing.data_generator.DataGenerator].

        !!! info
            More about Metalearners can be found here: [Metelearner][aucmedi.ensemble.metalearner]

            More about Aggregate functions can be found here: [aggregate][aucmedi.ensemble.aggregate]

        Args:
            prediction_generator (DataGenerator):   A data generator which will be used for inference.
            return_ensemble (bool):                 Option, whether gathered ensemble of predictions should be returned.

        Returns:
            preds (numpy.ndarray):                  A NumPy array of predictions formatted with shape (n_samples, n_labels).
            ensemble (numpy.ndarray):               Optional ensemble of predictions: Will be only passed if `return_ensemble=True`.
                                                    Shape (n_models, n_samples, n_labels).
        """
        # Verify if there is a linked cache dictionary
        con_tmp = (isinstance(self.cache_dir, tempfile.TemporaryDirectory) and \
                   os.path.exists(self.cache_dir.name))
        con_var = (self.cache_dir is not None and \
                   not isinstance(self.cache_dir, tempfile.TemporaryDirectory) \
                   and os.path.exists(self.cache_dir))
        if not con_tmp and not con_var:
            raise FileNotFoundError("Stacking does not have a valid model cache directory!")

        # Initialize some variables
        temp_dg = prediction_generator
        preds_ensemble = []
        preds_final = []

        # Extract data
        data_test = (temp_dg.samples, temp_dg.labels, temp_dg.metadata)

        # Identify path to model directory
        if isinstance(self.cache_dir, tempfile.TemporaryDirectory):
            path_model_dir = self.cache_dir.name
        else : path_model_dir = self.cache_dir

        # Sequentially iterate over model list
        for i in range(len(self.model_list)):
            path_model = os.path.join(path_model_dir,
                                      "nn_" + str(i) + ".model.keras")

            # Gather NeuralNetwork parameters
            model_paras = {
                "n_labels": self.model_list[i].n_labels,
                "channels": self.model_list[i].channels,
                "input_shape": self.model_list[i].input_shape,
                "architecture": self.model_list[i].architecture,
                "pretrained_weights": self.model_list[i].pretrained_weights,
                "loss": self.model_list[i].loss,
                "metrics": None,
                "activation_output": self.model_list[i].activation_output,
                "fcl_dropout": self.model_list[i].fcl_dropout,
                "meta_variables": self.model_list[i].meta_variables,
                "learning_rate": self.model_list[i].learning_rate,
            }

            # Gather DataGenerator parameters
            datagen_paras = {"path_imagedir": temp_dg.path_imagedir,
                             "batch_size": temp_dg.batch_size,
                             "data_aug": temp_dg.data_aug,
                             "seed": temp_dg.seed,
                             "subfunctions": temp_dg.subfunctions,
                             "shuffle": temp_dg.shuffle,
                             "standardize_mode": self.model_list[i].meta_standardize,
                             "resize": self.model_list[i].meta_input,
                             "grayscale": temp_dg.grayscale,
                             "prepare_images": temp_dg.prepare_images,
                             "sample_weights": temp_dg.sample_weights,
                             "image_format": temp_dg.image_format,
                             "loader": temp_dg.sample_loader,
                             "workers": temp_dg.workers,
                             "kwargs": temp_dg.kwargs
            }

            # Start inference process for model i
            process_queue = mp.Queue()
            process_pred = mp.Process(target=__prediction_process__,
                                      args=(process_queue,
                                            model_paras,
                                            path_model,
                                            data_test,
                                            datagen_paras))
            process_pred.start()
            process_pred.join()
            preds = process_queue.get()

            # Append preds to ensemble
            preds_ensemble.append(preds)

        # Preprocess prediction ensemble
        preds_ensemble = np.array(preds_ensemble)
        preds_ensemble = np.swapaxes(preds_ensemble, 0, 1)

        # Apply heterogenous metalearner
        if isinstance(self.ml_model, Metalearner_Base):
            s, m, c = preds_ensemble.shape
            x_stack = np.reshape(preds_ensemble, (s, m*c))
            preds_final = self.ml_model.predict(data=x_stack)
        # Apply homogeneous aggregate function
        elif isinstance(self.ml_model, Aggregate_Base):
            for i in range(preds_ensemble.shape[0]):
                pred_sample = self.ml_model.aggregate(preds_ensemble[i,:,:])
                preds_final.append(pred_sample)

        # Convert prediction list to NumPy
        preds_final = np.asarray(preds_final)

        # Return ensembled predictions
        if return_ensemble : return preds_final, np.swapaxes(preds_ensemble,1,0)
        else : return preds_final

    # Dump model to file
    def dump(self, directory_path):
        """ Store temporary Stacking models directory permanently to disk at desired location.

        If the model directory is a provided path which is already persistent on the disk,
        the directory is copied in order to keep original data persistent.

        Args:
            directory_path (str):       Path to store the model directory on disk.
        """
        if self.cache_dir is None:
            raise FileNotFoundError("Stacking does not have a valid model cache directory!")
        elif isinstance(self.cache_dir, tempfile.TemporaryDirectory):
            shutil.copytree(self.cache_dir.name, directory_path,
                            dirs_exist_ok=True)
            self.cache_dir.cleanup()
            self.cache_dir = directory_path
        else:
            shutil.copytree(self.cache_dir, directory_path, dirs_exist_ok=True)
            self.cache_dir = directory_path

    # Load model from file
    def load(self, directory_path):
        """ Load a Stacking model directory which can be used for Metalearner based inference.

        Args:
            directory_path (str):       Input path, from which the Stacking models will be loaded.
        """
        # Check directory existence
        if not os.path.exists(directory_path):
            raise FileNotFoundError("Provided model directory path does not exist!",
                                    directory_path)
        # Check model existence
        for i in range(len(self.model_list)):
            path_model = os.path.join(directory_path,
                                      "nn_" + str(i) + ".model.keras")
            if not os.path.exists(path_model):
                raise FileNotFoundError("Stacking model " + str(i) + \
                                        " does not exist!", path_model)
        # If heterogenous metalearner -> load metalearner model file
        if isinstance(self.ml_model, Metalearner_Base):
            path_model = os.path.join(directory_path,
                                      "metalearner.model.pickle")
            if not os.path.exists(path_model):
                raise FileNotFoundError("Metalearner model does not exist!",
                                        path_model)
            self.ml_model.load(path_model)

        # Update model directory
        self.cache_dir = directory_path

#-----------------------------------------------------#
#                     Subroutines                     #
#-----------------------------------------------------#
# Internal function for training a NeuralNetwork model in a separate process
def __training_process__(queue, model_paras, data_train, data_val,
                         datagen_paras, train_paras):
    # Extract data
    (train_x, train_y, train_m) = data_train
    (val_x, val_y, val_m) = data_val
    # Build training DataGenerator
    nn_train_gen = DataGenerator(train_x,
                                 path_imagedir=datagen_paras["path_imagedir"],
                                 labels=train_y,
                                 metadata=train_m,
                                 batch_size=datagen_paras["batch_size"],
                                 data_aug=datagen_paras["data_aug"],
                                 seed=datagen_paras["seed"],
                                 subfunctions=datagen_paras["subfunctions"],
                                 shuffle=datagen_paras["shuffle"],
                                 standardize_mode=datagen_paras["standardize_mode"],
                                 resize=datagen_paras["resize"],
                                 grayscale=datagen_paras["grayscale"],
                                 prepare_images=datagen_paras["prepare_images"],
                                 sample_weights=datagen_paras["sample_weights"],
                                 image_format=datagen_paras["image_format"],
                                 loader=datagen_paras["loader"],
                                 workers=datagen_paras["workers"],
                                 **datagen_paras["kwargs"])
    # Build validation DataGenerator
    nn_val_gen = DataGenerator(val_x,
                               path_imagedir=datagen_paras["path_imagedir"],
                               labels=val_y,
                               metadata=val_m,
                               batch_size=datagen_paras["batch_size"],
                               data_aug=None,
                               seed=datagen_paras["seed"],
                               subfunctions=datagen_paras["subfunctions"],
                               shuffle=False,
                               standardize_mode=datagen_paras["standardize_mode"],
                               resize=datagen_paras["resize"],
                               grayscale=datagen_paras["grayscale"],
                               prepare_images=datagen_paras["prepare_images"],
                               sample_weights=datagen_paras["sample_weights"],
                               image_format=datagen_paras["image_format"],
                               loader=datagen_paras["loader"],
                               workers=datagen_paras["workers"],
                               **datagen_paras["kwargs"])
    # Create NeuralNetwork
    model = NeuralNetwork(**model_paras)
    # Start NeuralNetwork training
    nn_history = model.train(nn_train_gen, nn_val_gen, **train_paras)
    # Store result in cache (which will be returned by the process queue)
    queue.put(nn_history)

# Internal function for inference with a fitted NeuralNetwork model in a separate process
def __prediction_process__(queue, model_paras, path_model, data_test,
                           datagen_paras):
    # Extract data
    (test_x, test_y, test_m) = data_test
    # Create inference DataGenerator
    nn_pred_gen = DataGenerator(test_x,
                                path_imagedir=datagen_paras["path_imagedir"],
                                labels=None,
                                metadata=test_m,
                                batch_size=datagen_paras["batch_size"],
                                data_aug=None,
                                seed=datagen_paras["seed"],
                                subfunctions=datagen_paras["subfunctions"],
                                shuffle=False,
                                standardize_mode=datagen_paras["standardize_mode"],
                                resize=datagen_paras["resize"],
                                grayscale=datagen_paras["grayscale"],
                                prepare_images=datagen_paras["prepare_images"],
                                sample_weights=datagen_paras["sample_weights"],
                                image_format=datagen_paras["image_format"],
                                loader=datagen_paras["loader"],
                                workers=datagen_paras["workers"],
                                **datagen_paras["kwargs"])
    # Create NeuralNetwork
    model = NeuralNetwork(**model_paras)
    # Load model weights from disk
    model.load(path_model)
    # Make prediction
    preds = model.predict(nn_pred_gen)
    # Store prediction results in cache (which will be returned by the process queue)
    queue.put(preds)



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# External libraries
import numpy as np
from PIL import Image
import matplotlib.cm as cm

#-----------------------------------------------------#
#                   Image Visualizer                  #
#-----------------------------------------------------#
def visualize_array(array, out_path=None):
    """ Simple wrapper function to visualize a NumPy matrix as image via PIL.

    ???+ info
        NumPy array shape has to be (x, y, channel) like this: (224, 224, 3)

    Args:
        array (numpy.ndarray):          NumPy matrix containing an image.
        out_path (str):                 Path in which image is stored (else live output).
    """
    # Ensure integer intensity values
    array = np.uint8(array)
    # Remove channel axis if grayscale
    if array.shape[-1] == 1 : array = np.reshape(array, array.shape[:-1])
    # Convert array to PIL image
    image = Image.fromarray(array)
    # Visualize or store image
    if out_path is None : image.show()
    else : image.save(out_path)

#-----------------------------------------------------#
#               XAI Heatmap Visualizer                #
#-----------------------------------------------------#
def visualize_heatmap(image, heatmap, out_path=None, alpha=0.4):
    """ Simple wrapper function to visualize a heatmap encoded as NumPy matrix with a
        [0-1] range as image via matplotlib and PILLOW.

    ??? abstract "Reference - Implementation"
        Author: François Chollet <br>
        Date: April 26, 2020 <br>
        https://keras.io/examples/vision/grad_cam/ <br>

    Args:
        image (numpy.ndarray):          NumPy matrix containing an image.
        heatmap (numpy.ndarray):        NumPy matrix containing a XAI heatmap.
        out_path (str):                 Path in which image is stored (else live output).
        alpha (float):                  Transparency value for heatmap overlap on image (range: [0-1]).
    """
    # If image is grayscale, convert to RGB
    if image.shape[-1] == 1 : image = np.concatenate((image,)*3, axis=-1)
    # Rescale heatmap to grayscale range
    heatmap = np.uint8(heatmap * 255)
    # Use jet colormap to colorize heatmap
    jet = cm.get_cmap("jet")
    # Use RGB values of the colormap
    jet_colors = jet(np.arange(256))[:,:3]
    jet_heatmap = jet_colors[heatmap] * 255
    # Superimpose the heatmap on original image
    si_img = jet_heatmap * alpha + (1-alpha) * image
    # Convert array to PIL image
    si_img = si_img.astype(np.uint8)
    pil_img = Image.fromarray(si_img)
    # Visualize or store image
    if out_path is None : pil_img.show()
    else : pil_img.save(out_path)



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# External libraries
import pickle
from sklearn.neighbors import KNeighborsClassifier
import numpy as np
# Internal libraries/scripts
from aucmedi.ensemble.metalearner.ml_base import Metalearner_Base

#-----------------------------------------------------#
#           Metalearner: k-Nearest Neighbors          #
#-----------------------------------------------------#
class KNearestNeighbors(Metalearner_Base):
    """ A k-Nearest Neighbors based Metalearner.

    This class should be passed to an ensemble function/class like Stacking for combining predictions.

    !!! info
        Can be utilized for binary, multi-class and multi-label tasks.

    ???+ abstract "Reference - Implementation"
        https://scikit-learn.org/stable/modules/generated/sklearn.neighbors.KNeighborsClassifier.html

        Scikit-learn: Machine Learning in Python, Pedregosa et al., JMLR 12, pp. 2825-2830, 2011.
        https://jmlr.csail.mit.edu/papers/v12/pedregosa11a.html
    """
    #---------------------------------------------#
    #                Initialization               #
    #---------------------------------------------#
    def __init__(self):
        self.model = KNeighborsClassifier()

    #---------------------------------------------#
    #                  Training                   #
    #---------------------------------------------#
    def train(self, x, y):
        # Train model
        self.model = self.model.fit(x, y)

    #---------------------------------------------#
    #                  Prediction                 #
    #---------------------------------------------#
    def predict(self, data):
        # Compute prediction probabilities via fitted model
        pred = self.model.predict_proba(data)
        # Postprocess decision tree predictions
        pred = np.asarray(pred)
        pred = np.swapaxes(pred[:,:,1], 0, 1)
        # Return results as NumPy array
        return pred

    #---------------------------------------------#
    #              Dump Model to Disk             #
    #---------------------------------------------#
    def dump(self, path):
        # Dump model to disk via pickle
        with open(path, "wb") as pickle_writer:
            pickle.dump(self.model, pickle_writer)

    #---------------------------------------------#
    #             Load Model from Disk            #
    #---------------------------------------------#
    def load(self, path):
        # Load model from disk via pickle
        with open(path, "rb") as pickle_reader:
            self.model = pickle.load(pickle_reader)



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                    Documentation                    #
#-----------------------------------------------------#
""" A pillar of any AUCMEDI pipeline: [aucmedi.neural_network.model.NeuralNetwork][]

The Neural Network class is a powerful interface to the deep neural network world in AUCMEDI.

???+ info "Pillars of AUCMEDI"
    - [aucmedi.data_processing.io_data.input_interface][]
    - [aucmedi.data_processing.data_generator.DataGenerator][]
    - [aucmedi.neural_network.model.NeuralNetwork][]

With an initialized Neural Network instance, it is possible to run training and predictions.

???+ example
    ```python
    # Import
    from aucmedi import *

    # Initialize model
    model = NeuralNetwork(n_labels=8, channels=3, architecture="2D.ResNet50")

    # Do some training
    datagen_train = DataGenerator(samples[:100], "images_dir/", labels=class_ohe[:100],
                                  resize=model.meta_input, standardize_mode=model.meta_standardize)
    model.train(datagen_train, epochs=50)

    # Do some predictions
    datagen_test = DataGenerator(samples[100:150], "images_dir/", labels=None,
                                 resize=model.meta_input, standardize_mode=model.meta_standardize)
    preds = model.predict(datagen_test)
    ```
"""



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
#External libraries
import unittest
import numpy as np
#Internal libraries
from aucmedi.data_processing.subfunctions import *

#-----------------------------------------------------#
#               Unittest: Subfunctions                #
#-----------------------------------------------------#
class SubfunctionsTEST(unittest.TestCase):
    # Create random imaging data
    @classmethod
    def setUpClass(self):
        np.random.seed(1234)
        # Create 2D Grayscale data
        img_gray = np.random.rand(16, 24, 1) * 255
        self.img2Dgray = np.float32(img_gray)
        # Create 2D RGB data
        img_rgb = np.random.rand(16, 24, 3) * 255
        self.img2Drgb = np.float32(img_rgb)
        # Create 3D Grayscale data
        img_3Dgray = np.random.rand(16, 24, 32, 1) * 255
        self.img3Dgray = np.float32(img_3Dgray)
        # Create 3D RGB data
        img_3Drgb = np.random.rand(16, 24, 32, 3) * 255
        self.img3Drgb = np.float32(img_3Drgb)
        # Create 3D HU data
        img_3Dhu = np.random.rand(16, 24, 32, 1) * 1500
        self.img3Dhu = np.float32(img_3Dhu - 500)

    #-------------------------------------------------#
    #              Subfunction: Padding               #
    #-------------------------------------------------#
    def test_PADDING_create(self):
        sf = Padding()

    def test_PADDING_transform(self):
        sf = Padding(shape=(32, 32), mode="edge")
        img_ppGRAY = sf.transform(self.img2Dgray.copy())
        self.assertTrue(np.array_equal(img_ppGRAY.shape, (32, 32, 1)))
        img_ppRGB = sf.transform(self.img2Drgb.copy())
        self.assertTrue(np.array_equal(img_ppRGB.shape, (32, 32, 3)))
        sf = Padding(shape=(8, 32), mode="constant")
        img_ppRGB = sf.transform(self.img2Drgb.copy())
        self.assertTrue(np.array_equal(img_ppRGB.shape, (16, 32, 3)))
        sf = Padding(mode="square")
        img_ppRGB = sf.transform(self.img2Drgb.copy())
        self.assertTrue(np.array_equal(img_ppRGB.shape, (24, 24, 3)))
        sf = Padding(mode="square")
        img_ppHU= sf.transform(self.img3Dhu.copy())
        self.assertTrue(np.array_equal(img_ppHU.shape, (32, 32, 32, 1)))
        sf = Padding(shape=(8, 32, 48), mode="edge")
        img_ppRGB = sf.transform(self.img3Drgb.copy())
        self.assertTrue(np.array_equal(img_ppRGB.shape, (16, 32, 48, 3)))

    #-------------------------------------------------#
    #               Subfunction: Resize               #
    #-------------------------------------------------#
    def test_RESIZE_create(self):
        sf = Resize()

    def test_RESIZE_transform(self):
        # 2D
        sf = Resize(shape=(32, 32))
        img_ppGRAY = sf.transform(self.img2Dgray.copy())
        self.assertTrue(np.array_equal(img_ppGRAY.shape, (32, 32, 1)))
        img_ppRGB = sf.transform(self.img2Drgb.copy())
        self.assertTrue(np.array_equal(img_ppRGB.shape, (32, 32, 3)))
        sf = Resize(shape=(8, 8))
        img_ppGRAY = sf.transform(self.img2Dgray.copy())
        self.assertTrue(np.array_equal(img_ppGRAY.shape, (8, 8, 1)))
        img_ppRGB = sf.transform(self.img2Drgb.copy())
        self.assertTrue(np.array_equal(img_ppRGB.shape, (8, 8, 3)))
        sf = Resize(shape=(32, 8))
        img_ppGRAY = sf.transform(self.img2Dgray.copy())
        self.assertTrue(np.array_equal(img_ppGRAY.shape, (32, 8, 1)))
        img_ppRGB = sf.transform(self.img2Drgb.copy())
        self.assertTrue(np.array_equal(img_ppRGB.shape, (32, 8, 3)))
        # 3D
        sf = Resize(shape=(32, 32, 32))
        img_ppGRAY = sf.transform(self.img3Dgray.copy())
        self.assertTrue(np.array_equal(img_ppGRAY.shape, (32, 32, 32, 1)))
        img_ppRGB = sf.transform(self.img3Drgb.copy())
        self.assertTrue(np.array_equal(img_ppRGB.shape, (32, 32, 32, 3)))
        sf = Resize(shape=(8, 8, 8))
        img_ppGRAY = sf.transform(self.img3Dgray.copy())
        self.assertTrue(np.array_equal(img_ppGRAY.shape, (8, 8, 8, 1)))
        img_ppRGB = sf.transform(self.img3Drgb.copy())
        self.assertTrue(np.array_equal(img_ppRGB.shape, (8, 8, 8, 3)))
        sf = Resize(shape=(32, 8, 8))
        img_ppGRAY = sf.transform(self.img3Dgray.copy())
        self.assertTrue(np.array_equal(img_ppGRAY.shape, (32, 8, 8, 1)))
        img_ppRGB = sf.transform(self.img3Drgb.copy())
        self.assertTrue(np.array_equal(img_ppRGB.shape, (32, 8, 8, 3)))

    #-------------------------------------------------#
    #             Subfunction: Standardize            #
    #-------------------------------------------------#
    def test_STANDARDIZE_create(self):
        sf = Standardize()

    def test_STANDARDIZE_transform(self):
        # Custom implementations
        for mode in ["z-score", "minmax", "grayscale"]:
            sf = Standardize(mode=mode)
            for data in [self.img2Dgray, self.img2Drgb, self.img3Dgray,
                         self.img3Drgb, self.img3Dhu]:
                img_pp = sf.transform(data.copy())
                if mode == "z-score":
                    self.assertTrue(np.amin(img_pp) <= 0)
                    self.assertTrue(np.amax(img_pp) >= 0)
                elif mode == "minmax":
                    self.assertTrue(np.amin(img_pp) >= 0)
                    self.assertTrue(np.amax(img_pp) <= 1)
                elif mode == "grayscale":
                    self.assertTrue(np.amin(img_pp) >= 0)
                    self.assertTrue(np.amax(img_pp) <= 255)
        # Keras implementations
        for mode in ["tf", "caffe", "torch"]:
            sf = Standardize(mode=mode)
            for data in [self.img2Drgb, self.img3Drgb]:
                img_pp = sf.transform(data.copy())
                if mode == "tf":
                    self.assertTrue(np.amin(img_pp) >= -1)
                    self.assertTrue(np.amax(img_pp) <= 1)
                elif mode == "caffe":
                    self.assertTrue(np.amin(img_pp) <= 0)
                    self.assertTrue(np.amax(img_pp) >= 0)
                elif mode == "torch":
                    self.assertTrue(np.amin(img_pp) <= 0)
                    self.assertTrue(np.amax(img_pp) >= 0)
            # self.assertRaises(ValueError, sf.transform, self.img3Dhu.copy())
        # Per channel normalization
        for mode in ["z-score", "minmax", "grayscale", "tf"]:
            sf = Standardize(mode=mode, per_channel=True)
            for data in [self.img2Drgb, self.img3Drgb]:
                img_pp = sf.transform(data.copy())
                for c in range(0, img_pp.shape[-1]):
                    if mode == "z-score":
                        self.assertTrue(np.amin(img_pp[..., c]) <= 0)
                        self.assertTrue(np.amax(img_pp[..., c]) >= 0)
                        if c != 0:
                            self.assertFalse(np.amax(img_pp[..., 0]) == \
                                             np.amax(img_pp[..., c]))
                    elif mode == "minmax":
                        self.assertTrue(np.amin(img_pp[..., c]) >= 0)
                        self.assertTrue(np.amax(img_pp[..., c]) <= 1)
                    elif mode == "grayscale":
                        self.assertTrue(np.amin(img_pp[..., c]) >= 0)
                        self.assertTrue(np.amax(img_pp[..., c]) <= 255)
                    elif mode == "tf":
                        self.assertTrue(np.amin(img_pp[..., c]) >= -1)
                        self.assertTrue(np.amax(img_pp[..., c]) <= 1)
                        if c != 0:
                            self.assertFalse(np.amax(img_pp[..., 0]) == \
                                             np.amax(img_pp[..., c]))

    #-------------------------------------------------#
    #              Subfunction: Cropping              #
    #-------------------------------------------------#
    def test_CROP_create(self):
        sf = Crop()
        sf = Crop(mode="center")
        sf = Crop(mode="random")
        self.assertRaises(ValueError, Crop, mode="test")

    def test_CROP_transform(self):
        # 2D testing
        sf = Crop(shape=(16, 12), mode="center")
        img_ppGRAY = sf.transform(self.img2Dgray.copy())
        self.assertTrue(np.array_equal(img_ppGRAY.shape, (16, 12, 1)))
        img_ppRGB = sf.transform(self.img2Drgb.copy())
        self.assertTrue(np.array_equal(img_ppRGB.shape, (16, 12, 3)))
        sf = Crop(shape=(8, 8), mode="random")
        img_ppGRAY = sf.transform(self.img2Dgray.copy())
        self.assertTrue(np.array_equal(img_ppGRAY.shape, (8, 8, 1)))
        img_ppRGB = sf.transform(self.img2Drgb.copy())
        self.assertTrue(np.array_equal(img_ppRGB.shape, (8, 8, 3)))
        # 3D testing
        sf = Crop(shape=(14, 12, 24))
        img_ppGRAY = sf.transform(self.img3Dgray.copy())
        self.assertTrue(np.array_equal(img_ppGRAY.shape, (14, 12, 24, 1)))
        img_ppRGB = sf.transform(self.img3Drgb.copy())
        self.assertTrue(np.array_equal(img_ppRGB.shape, (14, 12, 24, 3)))

    #-------------------------------------------------#
    #          Subfunction: Color Constancy           #
    #-------------------------------------------------#
    def test_COLORCONSTANCY_create(self):
        sf = ColorConstancy()

    def test_COLORCONSTANCY_transform(self):
        sf = ColorConstancy()
        img_filtered = sf.transform(self.img2Drgb.copy())
        self.assertFalse(np.array_equal(img_filtered, self.img2Drgb))
        self.assertTrue(np.array_equal(img_filtered.shape, (16, 24, 3)))
        img_filtered = sf.transform(self.img3Drgb.copy())
        self.assertFalse(np.array_equal(img_filtered, self.img3Drgb))
        self.assertTrue(np.array_equal(img_filtered.shape, (16, 24, 32, 3)))
        self.assertRaises(ValueError, sf.transform, self.img2Dgray.copy())

    #-------------------------------------------------#
    #                Subfunction: Clip                #
    #-------------------------------------------------#
    def test_CLIP_create(self):
        sf = Clip()

    def test_CLIP_transform(self):
        # global clipping
        sf = Clip(min=10)
        img_clipped = sf.transform(self.img3Dhu.copy())
        self.assertTrue(np.amin(img_clipped) >= 10)
        sf = Clip(max=30)
        img_clipped = sf.transform(self.img3Dhu.copy())
        self.assertTrue(np.amax(img_clipped) <= 30)
        sf = Clip(min=10, max=50)
        img_clipped = sf.transform(self.img3Dhu.copy())
        self.assertTrue(np.amin(img_clipped) >= 10)
        self.assertTrue(np.amax(img_clipped) <= 50)
        # per channel clipping
        sf = Clip(min=10, max=50, per_channel=True)
        img_clipped = sf.transform(self.img3Dgray.copy())
        self.assertTrue(np.amin(img_clipped[:,:,:,0]) >= 10)
        self.assertTrue(np.amax(img_clipped[:,:,:,0]) <= 50)
        sf = Clip(min=10, max=50, per_channel=True)
        img_clipped = sf.transform(self.img3Drgb.copy())
        for c in range(0,self.img3Drgb.shape[-1]):
            self.assertTrue(np.amin(img_clipped[:,:,:,c]) >= 10)
            self.assertTrue(np.amax(img_clipped[:,:,:,c]) <= 50)
        min_list = [10,20,60]
        max_list = [20,40,80]
        sf = Clip(min=min_list, max=max_list, per_channel=True)
        img_clipped = sf.transform(self.img2Drgb.copy())
        for c in range(0,self.img2Drgb.shape[-1]):
            self.assertTrue(np.amin(img_clipped[:,:,c]) >= min_list[c])
            self.assertTrue(np.amax(img_clipped[:,:,c]) <= max_list[c])
        sf = Clip(min=20, max=max_list, per_channel=True)
        img_clipped = sf.transform(self.img3Drgb.copy())
        for c in range(0,self.img3Drgb.shape[-1]):
            self.assertTrue(np.amin(img_clipped[:,:,:,c]) >= 20)
            self.assertTrue(np.amax(img_clipped[:,:,:,c]) <= max_list[c])
        sf = Clip(min=min_list, max=max_list, per_channel=True)
        img_clipped = sf.transform(self.img3Drgb.copy())
        for c in range(0,self.img3Drgb.shape[-1]):
            self.assertTrue(np.amin(img_clipped[:,:,:,c]) >= min_list[c])
            self.assertTrue(np.amax(img_clipped[:,:,:,c]) <= max_list[c])

    #-------------------------------------------------#
    #               Subfunction: Chromer              #
    #-------------------------------------------------#
    def test_CHROMER_create(self):
        sf = Chromer()
        sf = Chromer(target="grayscale")
        sf = Chromer(target="rgb")
        self.assertRaises(ValueError, Chromer, target="test")

    def test_CHROMER_transform(self):
        # Target Grayscale
        sf = Chromer(target="grayscale")
        img_filtered = sf.transform(self.img2Drgb.copy())
        self.assertFalse(np.array_equal(img_filtered, self.img2Drgb))
        self.assertTrue(np.array_equal(img_filtered.shape, (16, 24, 1)))
        img_filtered = sf.transform(self.img3Drgb.copy())
        self.assertFalse(np.array_equal(img_filtered, self.img3Drgb))
        self.assertTrue(np.array_equal(img_filtered.shape, (16, 24, 32, 1)))
        self.assertRaises(ValueError, sf.transform, self.img2Dgray.copy())
        # Target RGB
        sf = Chromer(target="rgb")
        img_filtered = sf.transform(self.img2Dgray.copy())
        self.assertFalse(np.array_equal(img_filtered, self.img2Dgray))
        self.assertTrue(np.array_equal(img_filtered.shape, (16, 24, 3)))
        img_filtered = sf.transform(self.img3Dgray.copy())
        self.assertFalse(np.array_equal(img_filtered, self.img3Dgray))
        self.assertTrue(np.array_equal(img_filtered.shape, (16, 24, 32, 3)))
        self.assertRaises(ValueError, sf.transform, self.img3Dhu.copy())
        self.assertRaises(ValueError, sf.transform, self.img2Drgb.copy())



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                    Documentation                    #
#-----------------------------------------------------#
""" Library of implemented Metalearners in AUCMEDI.

A Metalearner can be passed to an ensemble like Stacking and merges multiple class
predictions into a single prediction.

Metalearner are similar to [aggregate()][aucmedi.ensemble.aggregate] functions,
however Metalearners are models which require fitting before usage.

```
Assembled predictions encoded in a NumPy matrix with shape (N_models, N_classes).
Example: [[0.5, 0.4, 0.1],
          [0.4, 0.3, 0.3],
          [0.5, 0.2, 0.3]]
-> shape (3, 3)

Merged prediction encoded in a NumPy matrix with shape (1, N_classes).
Example: [[0.4, 0.3, 0.3]]
-> shape (1, 3)
```

!!! warning
    Not all Metalearners support multi-label classification!

    If a Metalearner has multi-label support can be found in its docs entry.

Metalearners are based on the abstract base class [Metalearner_Base][aucmedi.ensemble.metalearner.ml_base],
which allows simple integration of custom Metalearners for Ensemble.
"""
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# Import metalearners
from aucmedi.ensemble.metalearner.logistic_regression import LogisticRegression
from aucmedi.ensemble.metalearner.naive_bayes import NaiveBayes
from aucmedi.ensemble.metalearner.support_vector_machine import SupportVectorMachine
from aucmedi.ensemble.metalearner.gaussian_process import GaussianProcess
from aucmedi.ensemble.metalearner.decision_tree import DecisionTree
from aucmedi.ensemble.metalearner.best_model import BestModel
from aucmedi.ensemble.metalearner.averaging_mean_weighted import AveragingWeightedMean
from aucmedi.ensemble.metalearner.random_forest import RandomForest
from aucmedi.ensemble.metalearner.k_neighbors import KNearestNeighbors
from aucmedi.ensemble.metalearner.mlp import MLP

#-----------------------------------------------------#
#           Access Functions to Metalearners          #
#-----------------------------------------------------#
metalearner_dict = {"logistic_regression": LogisticRegression,
                    "naive_bayes": NaiveBayes,
                    "support_vector_machine": SupportVectorMachine,
                    "gaussian_process": GaussianProcess,
                    "decision_tree": DecisionTree,
                    "best_model": BestModel,
                    "weighted_mean": AveragingWeightedMean,
                    "random_forest": RandomForest,
                    "k_neighbors": KNearestNeighbors,
                    "mlp": MLP,
}
""" Dictionary of implemented Metalearners. """



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
#External libraries
import unittest
import numpy as np
#Internal libraries
from aucmedi.ensemble.aggregate import *

#-----------------------------------------------------#
#             Unittest: Aggregate Functions           #
#-----------------------------------------------------#
class AggregateTEST(unittest.TestCase):
    # Create random imaging and classification data
    @classmethod
    def setUpClass(self):
        np.random.seed(1234)
        # Create data
        self.pred_data = np.random.rand(10, 8)

    #-------------------------------------------------#
    #           Aggregate: Averaging by Mean          #
    #-------------------------------------------------#
    def test_Aggregate_Mean(self):
        agg_func = AveragingMean()
        self.assertTrue("mean" in aggregate_dict)
        agg_func = aggregate_dict["mean"]()
        pred = agg_func.aggregate(self.pred_data.copy())
        self.assertTrue(np.array_equal(pred.shape, (8,)))

    #-------------------------------------------------#
    #          Aggregate: Averaging by Median         #
    #-------------------------------------------------#
    def test_Aggregate_Median(self):
        agg_func = AveragingMedian()
        self.assertTrue("median" in aggregate_dict)
        agg_func = aggregate_dict["median"]()
        pred = agg_func.aggregate(self.pred_data.copy())
        self.assertTrue(np.array_equal(pred.shape, (8,)))

    #-------------------------------------------------#
    #         Aggregate: Majority Vote (Hard)         #
    #-------------------------------------------------#
    def test_Aggregate_MajorityVote(self):
        agg_func = MajorityVote()
        self.assertTrue("majority_vote" in aggregate_dict)
        agg_func = aggregate_dict["majority_vote"]()
        pred = agg_func.aggregate(self.pred_data.copy())
        self.assertTrue(np.array_equal(pred.shape, (8,)))

    #-------------------------------------------------#
    #                Aggregate: Softmax               #
    #-------------------------------------------------#
    def test_Aggregate_Softmax(self):
        agg_func = Softmax()
        self.assertTrue("softmax" in aggregate_dict)
        agg_func = aggregate_dict["softmax"]()
        pred = agg_func.aggregate(self.pred_data.copy())
        self.assertTrue(np.array_equal(pred.shape, (8,)))

    #-------------------------------------------------#
    #             Aggregate: Global Argmax            #
    #-------------------------------------------------#
    def test_Aggregate_GlobalArgmax(self):
        agg_func = GlobalArgmax()
        self.assertTrue("global_argmax" in aggregate_dict)
        agg_func = aggregate_dict["global_argmax"]()
        pred = agg_func.aggregate(self.pred_data.copy())
        self.assertTrue(np.array_equal(pred.shape, (8,)))



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# External libraries
import numpy as np
from sklearn.model_selection import StratifiedShuffleSplit, ShuffleSplit
# Internal libraries
from aucmedi.sampling.iterative import MultilabelStratifiedShuffleSplit

#-----------------------------------------------------#
#       Function: Sampling via Percentage Split       #
#-----------------------------------------------------#
def sampling_split(samples, labels, metadata=None, sampling=[0.8, 0.2],
                   stratified=True, iterative=False, seed=None):
    """ Simple wrapper function for calling percentage split sampling functions.

    Allow usage of stratified and iterative sampling algorithm.

    ???+ warning
        Be aware that multi-label data does not support random stratified sampling.

    Percentage split ratios have to be provided with a sampling list.
    Each percentage value in the list defines the approximate split size.
    Sum of percentage split ratios have to equal 1!

    ???+ example
        ```python title="Example for common train/val/test split"
        split_ratio = [0.7, 0.25, 0.05]
        ds = sampling_split(samples, labels, sampling=split_ratio)

        # Returns a list with the following elements as tuples:
        print(ds[0])  # -> (samples_a, labels_a)                with 70% of complete dataset
        print(ds[1])  # -> (samples_b, labels_b)                with 25% of complete dataset
        print(ds[2])  # -> (samples_c, labels_c)                with  5% of complete dataset
        ```

        ```python title="Example with metadata"
        ds = sampling_split(samples, labels, metadata, sampling=[0.8, 0.2])

        # Returns a list with the following elements as tuples:
        print(ds[0])  # -> (samples_a, labels_a, metadata_a)    with 80% of complete dataset
        print(ds[1])  # -> (samples_b, labels_b, metadata_b)    with 20% of complete dataset
        ```

    Args:
        samples (list of str):          List of sample/index encoded as Strings.
        labels (numpy.ndarray):         NumPy matrix containing the ohe encoded classification.
        metadata (numpy.ndarray):       NumPy matrix with additional metadata. Have to be shape (n_samples, meta_variables).
        sampling (list of float):       List of percentage values with split sizes.
        stratified (bool):              Option whether to use stratified sampling based on provided labels.
        iterative (bool):               Option whether to use iterative sampling algorithm.
        seed (int):                     Seed to ensure reproducibility for random functions.

    Returns:
        results (list of tuple):        List with `len(sampling)` containing tuples with sampled data:
                                        (samples_a, labels_a) and with metadata (samples_a, labels_a, metadata_a).
    """
    # Verify sampling percentages
    if not np.isclose(sum(sampling), 1.0):
        raise ValueError("Sum of Percentage split ratios as sampling do not" + \
                         " equal 1", sampling, np.sum(sampling))
    # Initialize leftover with the complete dataset
    leftover_samples = np.asarray(samples)
    leftover_labels = np.asarray(labels)
    if metadata is not None : leftover_meta = np.asarray(metadata)
    leftover_p = 0.0
    # Initialize result list
    results = []

    # Perform sampling for each percentage split
    for i in range(0, len(sampling)):
        # For last split, just take leftover data as subset
        if i == len(sampling)-1:
            # Generate split
            if metadata is None : split = (leftover_samples, leftover_labels)
            else : split = (leftover_samples, leftover_labels, leftover_meta)
            # Append splitted data and stop
            results.append(split)
            break

        # Identify split percentage for remaining data
        p = sampling[i] / (1.0 - leftover_p)
        # Initialize random sampler
        if not stratified and not iterative:
            sampler = ShuffleSplit(n_splits=1, random_state=seed,
                                   train_size=(1.0-p), test_size=p)
        # Initialize random stratified sampler
        elif stratified and not iterative:
            sampler = StratifiedShuffleSplit(n_splits=1, random_state=seed,
                                             train_size=(1.0-p), test_size=p)
        # Initialize iterative stratified sampler
        else:
            sampler = MultilabelStratifiedShuffleSplit(n_splits=1,
                            random_state=seed, train_size=(1.0-p), test_size=p)

        # Apply sampling
        subset_generator = sampler.split(X=leftover_samples, y=leftover_labels)
        subsets = next(subset_generator)
        # Generate split
        if metadata is None:
            split = (leftover_samples[subsets[1]], leftover_labels[subsets[1]])
        else : split = (leftover_samples[subsets[1]],
                        leftover_labels[subsets[1]],
                        leftover_meta[subsets[1]])
        # Append splitted data
        results.append(split)
        # Update remaining data
        leftover_p += sampling[i]
        leftover_samples = leftover_samples[subsets[0]]
        leftover_labels = leftover_labels[subsets[0]]
        if metadata is not None : leftover_meta = leftover_meta[subsets[0]]

    # Return result sampling
    return results



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
#External libraries
import unittest
import tempfile
import os
from PIL import Image
import numpy as np
#Internal libraries
from aucmedi.neural_network.loss_functions import *
from aucmedi import *

#-----------------------------------------------------#
#               Unittest: Loss Functions              #
#-----------------------------------------------------#
class LossfunctionsTEST(unittest.TestCase):
    # Create random imaging and classification data
    @classmethod
    def setUpClass(self):
        np.random.seed(1234)
        # Initialize temporary directory
        self.tmp_data = tempfile.TemporaryDirectory(prefix="tmp.aucmedi.",
                                                    suffix=".data")
        # Create data
        self.sampleList = []
        for i in range(0, 1):
            img = np.random.rand(32, 32, 3) * 255
            img_pillow = Image.fromarray(img.astype(np.uint8))
            index = "image.sample_" + str(i) + ".png"
            path_sample = os.path.join(self.tmp_data.name, index)
            img_pillow.save(path_sample)
            self.sampleList.append(index)

        # Create classification labels
        self.labels_ohe = np.zeros((1, 4), dtype=np.uint8)
        for i in range(0, 1):
            class_index = np.random.randint(0, 4)
            self.labels_ohe[i][class_index] = 1
        # Create Data Generator
        self.datagen = DataGenerator(self.sampleList, self.tmp_data.name,
                                     labels=self.labels_ohe, 
                                     resize=(32, 32),
                                     grayscale=False, batch_size=1)

    #-------------------------------------------------#
    #          Keras Categorical Crossentropy         #
    #-------------------------------------------------#
    def test_Keras(self):
        model = NeuralNetwork(n_labels=4, channels=3,
                               loss="categorical_crossentropy", 
                               input_shape=(32, 32))
        model.train(self.datagen, epochs=1)

    #-------------------------------------------------#
    #                Focal Loss: Binary               #
    #-------------------------------------------------#
    def test_FocalLoss_binary(self):
        lf = binary_focal_loss(alpha=0.25, gamma=2)
        model = NeuralNetwork(n_labels=4, channels=3, loss=lf, 
                                input_shape=(32, 32))
        model.train(self.datagen, epochs=1)

    #-------------------------------------------------#
    #             Focal Loss: Categorical             #
    #-------------------------------------------------#
    def test_FocalLoss_categorical(self):
        lf = categorical_focal_loss(alpha=[0.25, 0.25, 0.5, 4.0], gamma=2)
        model = NeuralNetwork(n_labels=4, channels=3, loss=lf, 
                                input_shape=(32, 32))
        model.train(self.datagen, epochs=1)

    #-------------------------------------------------#
    #             Focal Loss: Multi-Label             #
    #-------------------------------------------------#
    def test_FocalLoss_multilabel(self):
        lf = multilabel_focal_loss(class_weights=[0.25, 0.25, 0.5, 4.0], gamma=2)
        model = NeuralNetwork(n_labels=4, channels=3, loss=lf, 
                                input_shape=(32, 32))
        model.train(self.datagen, epochs=1)



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# External libraries
import numpy as np
from tensorflow.keras import backend as K
import tensorflow as tf

#-----------------------------------------------------#
#                 Focal Loss - Binary                 #
#-----------------------------------------------------#
def binary_focal_loss(alpha=0.25, gamma=2.0):
    """ Binary form of focal loss computation.

    FL(p_t) = -alpha * (1 - p_t)**gamma * log(p_t)
    where p = sigmoid(x), p_t = p or 1 - p depending on if the label is 1 or 0, respectively.

    ??? example
        ```python
        from aucmedi.neural_network.loss_functions import *
        my_loss = binary_focal_loss(alpha=0.75)

        model = NeuralNetwork(n_labels=1, channels=3, loss=my_loss)
        ```

    ??? abstract "Reference - Implementation"
        Author: Umberto Griffo <br>
        GitHub: [https://github.com/umbertogriffo](https://github.com/umbertogriffo) <br>
        Source: [https://github.com/umbertogriffo/focal-loss-keras](https://github.com/umbertogriffo/focal-loss-keras) <br>

    ??? abstract "Reference - Publication"
        Focal Loss for Dense Object Detection (Aug 2017) <br>
        Authors: Tsung-Yi Lin, Priya Goyal, Ross Girshick, Kaiming He, Piotr Dollár <br>
        [https://arxiv.org/abs/1708.02002](https://arxiv.org/abs/1708.02002)

    Args:
        alpha (float):      Class weight for positive class.
        gamma (float):      Tunable focusing parameter (γ ≥ 0).

    Returns:
        loss (Loss Function):               A TensorFlow compatible loss function. This object can be
                                            passed to the [NeuralNetwork][aucmedi.neural_network.model.NeuralNetwork] `loss` parameter.
    """
    def binary_focal_loss_fixed(y_true, y_pred):
        y_true = tf.cast(y_true, tf.float32)
        # Define epsilon so that the back-propagation will not result in NaN for 0 divisor case
        epsilon = K.epsilon()
        # Add the epsilon to prediction value
        # y_pred = y_pred + epsilon
        # Clip the prediciton value
        y_pred = K.clip(y_pred, epsilon, 1.0 - epsilon)
        # Calculate p_t
        p_t = tf.where(K.equal(y_true, 1), y_pred, 1 - y_pred)
        # Calculate alpha_t
        alpha_factor = K.ones_like(y_true) * alpha
        alpha_t = tf.where(K.equal(y_true, 1), alpha_factor, 1 - alpha_factor)
        # Calculate cross entropy
        cross_entropy = -K.log(p_t)
        weight = alpha_t * K.pow((1 - p_t), gamma)
        # Calculate focal loss
        loss = weight * cross_entropy
        # Sum the losses in mini_batch
        loss = K.mean(K.sum(loss, axis=1))
        return loss

    return binary_focal_loss_fixed

#-----------------------------------------------------#
#              Focal Loss - Categorical               #
#-----------------------------------------------------#
def categorical_focal_loss(alpha, gamma=2.0):
    """ Softmax version of focal loss.

    When there is a skew between different categories/labels in your data set,
    you can try to apply this function as a loss.

    ```
           m
      FL = ∑  -alpha * (1 - p_o,c)^gamma * y_o,c * log(p_o,c)
          c=1

      where m = number of classes, c = class and o = observation
    ```

    The `class_weights_list` obtained from [compute_class_weights][aucmedi.utils.class_weights.compute_class_weights]
    can be provided as parameter `alpha`.

    ??? example
        ```python
        # Compute class weights
        from aucmedi.utils.class_weights import compute_class_weights
        cw_loss, cw_fit = compute_class_weights(class_ohe)

        from aucmedi.neural_network.loss_functions import *
        my_loss = categorical_focal_loss(alpha=cw_loss)

        model = NeuralNetwork(n_labels=6, channels=3, loss=my_loss)
        ```

    ??? abstract "Reference - Implementation"
        Author: Umberto Griffo <br>
        GitHub: [https://github.com/umbertogriffo](https://github.com/umbertogriffo) <br>
        Source: [https://github.com/umbertogriffo/focal-loss-keras](https://github.com/umbertogriffo/focal-loss-keras) <br>

    ??? abstract "Reference - Publication"
        Focal Loss for Dense Object Detection (Aug 2017) <br>
        Authors: Tsung-Yi Lin, Priya Goyal, Ross Girshick, Kaiming He, Piotr Dollár <br>
        [https://arxiv.org/abs/1708.02002](https://arxiv.org/abs/1708.02002)

    Args:
        alpha (list of float):      The same as weighing factor in balanced cross entropy.
                                    Alpha is used to specify the weight of different categories/labels,
                                    the size of the array needs to be consistent with the number of classes.
        gamma (float):              Focusing parameter for modulating factor (1-p).

    Returns:
        loss (Loss Function):               A TensorFlow compatible loss function. This object can be
                                            passed to the [NeuralNetwork][aucmedi.neural_network.model.NeuralNetwork] `loss` parameter.
    """
    alpha = np.array(alpha, dtype=np.float32)

    def categorical_focal_loss_fixed(y_true, y_pred):
        y_true = tf.cast(y_true, tf.float32)
        # Clip the prediction value to prevent NaN's and Inf's
        epsilon = K.epsilon()
        y_pred = K.clip(y_pred, epsilon, 1. - epsilon)

        # Calculate Cross Entropy
        cross_entropy = -y_true * K.log(y_pred)

        # Calculate Focal Loss
        loss = alpha * K.pow(1 - y_pred, gamma) * cross_entropy

        # Compute mean loss in mini_batch
        return K.mean(K.sum(loss, axis=-1))

    return categorical_focal_loss_fixed

#-----------------------------------------------------#
#               Focal Loss - Multilabel               #
#-----------------------------------------------------#
def multilabel_focal_loss(class_weights, gamma=2.0,
                          class_sparsity_coefficient=1.0):
    """ Focal loss for multi-label classification.

    ??? example
        ```python
        # Compute class weights
        from aucmedi.utils.class_weights import compute_class_weights
        class_weights = compute_multilabel_weights(class_ohe)

        from aucmedi.neural_network.loss_functions import *
        my_loss = multilabel_focal_loss(class_weights=class_weights)

        model = NeuralNetwork(n_labels=6, channels=3, loss=my_loss,
                               activation_output="sigmoid")
        ```

    ??? abstract "Reference - Implementation"
        Author: Sushant Tripathy <br>
        LinkedIn: [https://www.linkedin.com/in/sushanttripathy/](https://www.linkedin.com/in/sushanttripathy/) <br>
        Source: [https://github.com/sushanttripathy/Keras_loss_functions/blob/master/focal_loss.py](https://github.com/sushanttripathy/Keras_loss_functions/blob/master/focal_loss.py) <br>

    ??? abstract "Reference - Publication"
        Focal Loss for Dense Object Detection (Aug 2017) <br>
        Authors: Tsung-Yi Lin, Priya Goyal, Ross Girshick, Kaiming He, Piotr Dollár  <br>
        [https://arxiv.org/abs/1708.02002](https://arxiv.org/abs/1708.02002) <br>

    Args:
        class_weights (list of float):      Non-zero, positive class-weights. This is used instead
                                            of alpha parameter.
        gamma (float):                      The Gamma parameter in Focal Loss. Default value (2.0).
        class_sparsity_coefficient (float): The weight of True labels over False labels. Useful
                                            if True labels are sparse. Default value (1.0).
    Returns:
        loss (Loss Function):               A TensorFlow compatible loss function. This object can be
                                            passed to the [NeuralNetwork][aucmedi.neural_network.model.NeuralNetwork] `loss` parameter.
    """
    class_weights = K.constant(class_weights, tf.float32)
    gamma = K.constant(gamma, tf.float32)
    class_sparsity_coefficient = K.constant(class_sparsity_coefficient,
                                            tf.float32)

    def focal_loss_function(y_true, y_pred):
        y_true = tf.cast(y_true, tf.float32)

        predictions_0 = (1.0 - y_true) * y_pred
        predictions_1 = y_true * y_pred

        cross_entropy_0 = (1.0 - y_true) * (-K.log(K.clip(1.0 - predictions_0,
                                K.epsilon(), 1.0 - K.epsilon())))
        cross_entropy_1 = y_true *(class_sparsity_coefficient * -K.log(K.clip(
                                predictions_1, K.epsilon(), 1.0 - K.epsilon())))

        cross_entropy = cross_entropy_1 + cross_entropy_0
        class_weighted_cross_entropy = cross_entropy * class_weights

        weight_1 = K.pow(K.clip(1.0 - predictions_1,
                                K.epsilon(), 1.0 - K.epsilon()), gamma)
        weight_0 = K.pow(K.clip(predictions_0, K.epsilon(),
                                1.0 - K.epsilon()), gamma)

        weight = weight_0 + weight_1
        focal_loss_tensor = weight * class_weighted_cross_entropy

        return K.mean(focal_loss_tensor, axis=1)

    return focal_loss_function



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# External Libraries
import numpy as np
import pandas as pd
import os
from plotnine import *
# Internal libraries/scripts

#-----------------------------------------------------#
#            Evaluation - Dataset Analysis            #
#-----------------------------------------------------#
def evaluate_dataset(samples,
                     labels,
                     out_path,
                     class_names=None,
                     show=False,
                     plot_barplot=False,
                     plot_heatmap=False,
                     suffix=None):
    """ Function for dataset evaluation (descriptive statistics).

    ???+ example
        ```python
        # Import libraries
        from aucmedi import *
        from aucmedi.evaluation import *

        # Peak data information via the first pillar of AUCMEDI
        ds = input_interface(interface="csv",                       # Interface type
                             path_imagedir="dataset/images/",
                             path_data="dataset/annotations.csv",
                             ohe=False, col_sample="ID", col_class="diagnosis")
        (samples, class_ohe, nclasses, class_names, image_format) = ds

        # Pass information to the evaluation function
        evaluate_dataset(samples, class_ohe, out_path="./", class_names=class_names)
        ```

    Created files in directory of `out_path`:

    - "plot.dataset.barplot.png"
    - "plot.dataset.heatmap.png"

    ???+ info "Preview for Bar Plot"
        ![Evaluation_Dataset_Barplot](../../images/evaluation.plot.dataset.barplot.png)

        Based on dataset: [ISIC 2019 Challenge](https://challenge.isic-archive.com/landing/2019/).

    ???+ info "Preview for Heatmap"
        ![Evaluation_Dataset_Heatmap](../../images/evaluation.plot.dataset.heatmap.png)

        Based on first 50 samples from dataset: [ISIC 2019 Challenge](https://challenge.isic-archive.com/landing/2019/).

    Args:
        samples (list of str):              List of sample/index encoded as Strings. Provided by
                                            [input_interface][aucmedi.data_processing.io_data.input_interface].
        labels (numpy.ndarray):             Classification list with One-Hot Encoding. Provided by
                                            [input_interface][aucmedi.data_processing.io_data.input_interface].
        out_path (str):                     Path to directory in which plotted figures are stored.
        class_names (list of str):          List of names for corresponding classes. Used for evaluation. Provided by
                                            [input_interface][aucmedi.data_processing.io_data.input_interface].
                                            If not provided (`None` provided), class indices will be used.
        show (bool):                        Option, whether to also display the generated charts.
        plot_barplot (bool):                Option, whether to generate a bar plot of class distribution.
        plot_heatmap (bool):                Option, whether to generate a heatmap of class overview. Only recommended for subsets of ~50 samples.
        suffix (str):                       Special suffix to add in the created figure filename.

    Returns:
        df_cf (pandas.DataFrame):           Dataframe containing the class distribution of the dataset.
    """

    # Generate barplot
    df_cf = evalby_barplot(labels, out_path, class_names, plot_barplot, show,
                           suffix)

    # Generate heatmap
    if plot_heatmap:
        evalby_heatmap(samples, labels, out_path, class_names, show, suffix)

    # Return table with class distribution
    return df_cf

#-----------------------------------------------------#
#             Dataset Analysis - Barplot              #
#-----------------------------------------------------#
def evalby_barplot(labels, out_path, class_names, plot_barplot, show=False,
                   suffix=None):
    # compute class frequency
    cf_list = []
    for c in range(0, labels.shape[1]):
        n_samples = labels.shape[0]
        class_freq = np.sum(labels[:, c])
        if class_names is None : curr_class = str(c)
        else : curr_class = class_names[c]
        class_percentage = np.round(class_freq / n_samples, 2) * 100
        cf_list.append([curr_class, class_freq, class_percentage])

    # Convert class frequency results to dataframe
    df_cf = pd.DataFrame(np.array(cf_list),
                         columns=["class", "class_freq", "class_perc"])
    df_cf["class_perc"] = pd.to_numeric(df_cf["class_perc"])

    if plot_barplot:
        # Plot class frequency results
        fig = (ggplot(df_cf, aes("class", "class_perc", fill="class"))
                   + geom_bar(stat="identity", color="black")
                   + geom_text(aes(label="class_freq"), nudge_y=5)
                   + coord_flip()
                   + ggtitle("Dataset Analysis: Class Distribution")
                   + xlab("Classes")
                   + ylab("Class Frequency (in %)")
                   + scale_y_continuous(limits=[0, 100],
                                        breaks=np.arange(0,110,10))
                   + theme_bw()
                   + theme(legend_position="none"))

        # Store figure to disk
        filename = "plot.dataset.barplot"
        if suffix is not None : filename += "." + str(suffix)
        filename += ".png"
        fig.save(filename=filename, path=out_path, width=10, height=9, dpi=200)

        # Plot figure
        if show : print(fig)

    # Return class table
    return df_cf

#-----------------------------------------------------#
#             Dataset Analysis - Heatmap              #
#-----------------------------------------------------#
def evalby_heatmap(samples, labels, out_path, class_names, show=False,
                   suffix=None):
    # Create dataframe
    if class_names is None : df = pd.DataFrame(labels, index=samples)
    else : df = pd.DataFrame(labels, index=samples, columns=class_names)

    # Preprocess dataframe
    df = df.reset_index()
    df_melted = pd.melt(df, id_vars="index", var_name="class",
                        value_name="presence")

    # Plot heatmap
    fig = (ggplot(df_melted, aes("index", "class", fill="presence"))
               + geom_tile()
               + coord_flip()
               + ggtitle("Dataset Analysis: Overview")
               + xlab("Samples")
               + ylab("Classes")
               + scale_fill_gradient(low="white", high="#3399FF")
               + theme_classic()
               + theme(legend_position="none"))

    # Store figure to disk
    filename = "plot.dataset.heatmap"
    if suffix is not None : filename += "." + str(suffix)
    filename += ".png"
    fig.save(filename=filename, path=out_path, width=10, height=9, dpi=200)

    # Plot figure
    if show : print(fig)



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                    Documentation                    #
#-----------------------------------------------------#
""" The IO Interfaces of AUCMEDI allow extracting information from datasets in different
    structures or formats.

These interfaces are called **internally** via the [input_interface][aucmedi.data_processing.io_data.input_interface].

???+ info "Format Interfaces"
    | Interface      | Internal Function                                                    | Description                                  |
    | -------------- | -------------------------------------------------------------------- | -------------------------------------------- |
    |  `"csv"`       | [io_csv()][aucmedi.data_processing.io_interfaces.io_csv]             | Storing class annotations in a CSV file.     |
    |  `"directory"` | [io_directory()][aucmedi.data_processing.io_interfaces.io_directory] | Storing class annotations in subdirectories. |
    |  `"json"`      | [io_json()][aucmedi.data_processing.io_interfaces.io_json]           | Storing class annotations in a JSON file.    |

"""
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
from aucmedi.data_processing.io_interfaces.io_csv import csv_loader
from aucmedi.data_processing.io_interfaces.io_json import json_loader
from aucmedi.data_processing.io_interfaces.io_directory import directory_loader



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# External libraries
import os
import numpy as np
from PIL import Image

#-----------------------------------------------------#
#             Image Loader for AUCMEDI IO             #
#-----------------------------------------------------#
def image_loader(sample, path_imagedir, image_format=None, grayscale=False,
                 **kwargs):
    """ Image Loader for image loading within the AUCMEDI pipeline.

    The Image Loader is an IO_loader function, which have to be passed to the
    [DataGenerator][aucmedi.data_processing.data_generator.DataGenerator].

    ???+ info
        The Image Loader utilizes Pillow for image loading: <br>
        https://github.com/python-pillow/Pillow

    ???+ example
        ```python
        # Import required libraries
        from aucmedi import *

        # Initialize input data reader
        ds = input_interface(interface="csv",
                             path_imagedir="dataset/images/",
                             path_data="dataset/annotations.csv",
                             ohe=False, col_sample="ID", col_class="diagnosis")
        (samples, class_ohe, nclasses, class_names, image_format) = ds

        # Initialize DataGenerator with by default using image_loader
        data_gen = DataGenerator(samples, "dataset/images/", labels=class_ohe,
                                 image_format=image_format, resize=None)

        # Initialize DataGenerator with manually selected image_loader
        from aucmedi.data_processing.io_loader import image_loader
        data_gen = DataGenerator(samples, "dataset/images/", labels=class_ohe,
                                 image_format=image_format, resize=None,
                                 loader=image_loader)
        ```

    Args:
        sample (str):               Sample name/index of an image.
        path_imagedir (str):        Path to the directory containing the images.
        image_format (str):         Image format to add at the end of the sample index for image loading.
        grayscale (bool):           Boolean, whether images are grayscale or RGB.
        **kwargs (dict):            Additional parameters for the sample loader.
    """
    # Get image path
    if image_format : img_file = sample + "." + image_format
    else : img_file = sample
    path_img = os.path.join(path_imagedir, img_file)
    # Load image via the PIL package
    img_raw = Image.open(path_img)
    # Convert image to grayscale or rgb
    if grayscale : img_converted = img_raw.convert('LA')
    else : img_converted = img_raw.convert('RGB')
    # Convert image to NumPy
    img = np.asarray(img_converted)
    # Perform additional preprocessing if grayscale image
    if grayscale:
        # Remove maximum value and keep only intensity
        img = img[:,:,0]
        # Reshape image to create a single channel
        img = np.reshape(img, img.shape + (1,))
    # Return image
    return img



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# External libraries
import numpy as np
# Internal libraries/scripts
from aucmedi.data_processing.subfunctions.sf_base import Subfunction_Base

#-----------------------------------------------------#
#              Subfunction class: Chromer             #
#-----------------------------------------------------#
""" A chromer Subfunction class which can be used for transforming:
        grayscale ->  RGB
        RGB       ->  grayscale

    Possible target formats : ["grayscale", "rgb"]

    Typical use case is converting a grayscale to RGB in order to utilize
    transfer learning weights on ImageNet.

Methods:
    __init__                Object creation function.
    transform:              Apply chromer.
"""
class Chromer(Subfunction_Base):
    """ A Subfunction class which which can be used for color format transforming.

    ```
    Transforms:
    - grayscale ->  RGB
    - RGB       ->  grayscale

    ```

    Possible target formats: `["rgb", "grayscale"]`

    Typical use case is converting a grayscale to RGB in order to utilize
    transfer learning weights based on ImageNet.
    """
    #---------------------------------------------#
    #                Initialization               #
    #---------------------------------------------#
    def __init__(self, target="rgb"):
        """ Initialization function for creating a Chromer Subfunction which can be passed to a
            [DataGenerator][aucmedi.data_processing.data_generator.DataGenerator].

        Args:
            target (str):               Transformation mode for desired target format.
        """
        # Verify target format
        if target not in ["grayscale", "rgb"]:
            raise ValueError("Unknown target format for Chromer Subfunction",
                             target, "Possibles target formats are: ['grayscale', 'rgb']")
        # Cache target format
        self.target = target

    #---------------------------------------------#
    #                Transformation               #
    #---------------------------------------------#
    def transform(self, image):
        # Verify that image is in correct format
        if self.target == "rgb" and (image.shape[-1] != 1 or \
                                     np.max(image) > 255 or \
                                     np.min(image) < 0):
            raise ValueError("Subfunction Chromer: Image is not in grayscale format!",
                             "Ensure that it is grayscale normalized and has",
                             "a single channel.")
        elif self.target == "grayscale" and (image.shape[-1] != 3 or \
                                             np.max(image) > 255 or \
                                             np.min(image) < 0):
            raise ValueError("Subfunction Chromer: Image is not in RGB format!",
                             "Ensure that it is normalized [0,255] and has 3 channels.")
        # Run grayscale -> RGB
        if self.target == "rgb":
            image_chromed = np.concatenate((image,)*3, axis=-1)
        # Run RGB -> grayscale
        else:
            # Get color intensity values
            r = np.take(image, indices=0, axis=-1)
            g = np.take(image, indices=1, axis=-1)
            b = np.take(image, indices=2, axis=-1)
            # Compute grayscale image
            image_chromed = 0.299 * r + 0.587 * g + 0.114 * b
            # Add channel axis back
            image_chromed = np.expand_dims(image_chromed, axis=-1)
        # Return chromed image
        return image_chromed


# convert grayscale to RGB
# convert RGB to grayscale



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# External Libraries
import numpy as np
import pandas as pd
import os
from plotnine import *
# Internal libraries/scripts
from aucmedi.evaluation.metrics import *

#-----------------------------------------------------#
#           Evaluation - Compare Performance          #
#-----------------------------------------------------#
def evaluate_comparison(pred_list,
                        labels,
                        out_path,
                        model_names=None,
                        class_names=None,
                        multi_label=False,
                        metrics_threshold=0.5,
                        macro_average_classes=False,
                        suffix=None):
    """ Function for performance comparison evaluation based on predictions from multiple models.

    ???+ example
        ```python
        # Import libraries
        from aucmedi import *
        from aucmedi.evaluation import *
        from aucmedi.ensemble import *

        # Load data
        ds = input_interface(interface="csv",                       # Interface type
                             path_imagedir="dataset/images/",
                             path_data="dataset/annotations.csv",
                             ohe=False, col_sample="ID", col_class="diagnosis")
        (samples, class_ohe, nclasses, class_names, image_format) = ds

        # Initialize model
        model_a = NeuralNetwork(n_labels=8, channels=3, architecture="2D.ResNet50")

        # Initialize Bagging object for 3-fold cross-validation
        el = Bagging(model_a, k_fold=3)

        # Do some predictions via Bagging (return also all prediction ensembles)
        datagen_test = DataGenerator(samples, "dataset/images/", labels=None,
                                     resize=model.meta_input, standardize_mode=model.meta_standardize)
        pred_merged, pred_ensemble = el.predict(datagen_test, return_ensemble=True)

        # Pass prediction ensemble to evaluation function
        evaluate_comparison(pred_ensemble, class_ohe, out_path="./", class_names=class_names)


        # Do some predictions with manually initialized models
        model_b = NeuralNetwork(n_labels=8, channels=3, architecture="2D.DenseNet121")
        model_c = NeuralNetwork(n_labels=8, channels=3, architecture="2D.MobileNetV2")

        pred_a = model_a.predict(datagen_test)
        pred_b = model_b.predict(datagen_test)
        pred_c = model_c.predict(datagen_test)

        pred_ensemble = [pred_a, pred_b, pred_c]

        # Pass prediction ensemble to evaluation function
        evaluate_comparison(pred_ensemble, class_ohe, out_path="./", class_names=class_names)
        ```

    Created files in directory of `out_path`:

    - "plot.comparison.beside.png"
    - "plot.comparison.gain.png"

    ???+ info "Preview for Bar Plot"
        ![Evaluation_Comparison_Beside](../../images/evaluation.plot.comparison.beside.png)

        Predictions based on [ISIC 2019 Challenge](https://challenge.isic-archive.com/landing/2019/) with
        macro-averaged class-wise metrics.

    ???+ info "Preview for Confusion Matrix"
        ![Evaluation_Comparison_Gain](../../images/evaluation.plot.comparison.gain.png)

        Predictions based on [ISIC 2019 Challenge](https://challenge.isic-archive.com/landing/2019/).

    Args:
        pred_list (list of numpy.ndarray):  A list of NumPy arrays containing predictions from multiple models formatted with shape
                                            (n_models, n_samples, n_labels). Provided by [NeuralNetwork][aucmedi.neural_network.model].
        labels (numpy.ndarray):             Classification list with One-Hot Encoding. Provided by
                                            [input_interface][aucmedi.data_processing.io_data.input_interface].
        out_path (str):                     Path to directory in which plotted figures are stored.
        model_names (list of str):          List of names for corresponding models which are for visualization. If not provided (`None`
                                            provided), model index of `pred_list` will be used.
        class_names (list of str):          List of names for corresponding classes. Used for evaluation. Provided by
                                            [input_interface][aucmedi.data_processing.io_data.input_interface].
                                            If not provided (`None` provided), class indices will be used.
        multi_label (bool):                 Option, whether task is multi-label based (has impact on evaluation).
        metrics_threshold (float):          Only required if 'multi_label==True`. Threshold value if prediction is positive.
                                            Used in metric computation for CSV and bar plot.
        macro_average_classes (bool):       Option, whether classes should be macro-averaged in order to increase visualization overview.
        suffix (str):                       Special suffix to add in the created figure filename.

    Returns:
        df_merged (pandas.DataFrame):       Dataframe containing the merged metrics of all models.
        df_gain (pandas.DataFrame):         Dataframe containing performance gain compared to first model.
    """
    # Identify number of labels
    n_labels = labels.shape[-1]
    # Identify prediction threshold
    if multi_label : threshold = metrics_threshold
    else : threshold = None

    # Compute metric dataframe for each mode
    df_list = []
    for m in range(0, len(pred_list)):
        metrics = compute_metrics(pred_list[m], labels, n_labels, threshold)

        # Rename class association in metrics dataframe
        class_mapping = {}
        if class_names is not None:
            for c in range(len(class_names)):
                class_mapping[c] = class_names[c]
            metrics["class"].replace(class_mapping, inplace=True)
        if class_names is None:
            metrics["class"] = pd.Categorical(metrics["class"])

        # Assign model name to dataframe
        if model_names is not None : metrics["model"] = model_names[m]
        else : metrics["model"] = "model_" + str(m)

        # Optional: Macro average classes
        if macro_average_classes:
            metrics_avg = metrics.groupby(["metric", "model"])[["score"]].mean()
            metrics = metrics_avg.reset_index()

        # Append to dataframe list
        df_list.append(metrics)
    # Combine dataframes
    df_merged = pd.concat(df_list, axis=0, ignore_index=True)

    # Generate comparison beside plot
    evalby_beside(df_merged, out_path, suffix)

    # Generate comparison gain plot
    df_gain = evalby_gain(df_merged, out_path, suffix)

    # Return combined and gain dataframe
    return df_merged, df_gain

#-----------------------------------------------------#
#           Evaluation Comparison - Beside            #
#-----------------------------------------------------#
def evalby_beside(df, out_path, suffix=None):
    # Remove confusion matrix from dataframe
    df = df[~df["metric"].isin(["TN", "FN", "FP", "TP"])]

    # Plot metric results class-wise
    if "class" in df.columns:
        fig = (ggplot(df, aes("model", "score", fill="model"))
                  + geom_col(stat='identity', width=0.6, color="black",
                             position = position_dodge(width=0.6))
                  + ggtitle("Performance Comparison: Metric Overview")
                  + facet_grid("metric ~ class")
                  + coord_flip()
                  + xlab("")
                  + ylab("Score")
                  + scale_y_continuous(limits=[0, 1])
                  + theme_bw()
                  + theme(legend_position="none"))
    # Plot metric results class macro-averaged
    else:
        fig = (ggplot(df, aes("model", "score", fill="model"))
                  + geom_col(stat='identity', width=0.6, color="black",
                             position = position_dodge(width=0.6))
                  + ggtitle("Performance Comparison: Metric Overview")
                  + facet_wrap("metric")
                  + coord_flip()
                  + xlab("")
                  + ylab("Score")
                  + scale_y_continuous(limits=[0, 1])
                  + theme_bw()
                  + theme(legend_position="none"))

    # Store figure to disk
    filename = "plot.comparison.beside"
    if suffix is not None : filename += "." + str(suffix)
    filename += ".png"
    fig.save(filename=filename, path=out_path, width=18, height=9, dpi=300)

#-----------------------------------------------------#
#            Evaluation Comparison - Gain             #
#-----------------------------------------------------#
def evalby_gain(df, out_path, suffix=None):
    # Remove confusion matrix from dataframe
    df = df[~df["metric"].isin(["TN", "FN", "FP", "TP"])]

    # Define gain computation function
    def compute_gain(row, template):
        # Identify current metric
        m = row["metric"]
        # Obtain class-wise divisor
        if "class" in row.index:
            c = row["class"]
            div = template.loc[(template["metric"] == m) & \
                               (template["class"] == c)]["score"].values[0]
        # Obtain macro-averaged divisor
        else:
            div = template.loc[template["metric"] == m]["score"].values[0]
        # Compute gain in percentage compared to template model
        row["score"] = (row["score"] / div) - 1.0
        return row

    # Compute percentage gain compared to first model
    first_model = df["model"].iloc[0]
    template = df.loc[df["model"] == first_model]
    df = df.apply(compute_gain, axis=1, args=(template,))

    # Plot gain results class-wise
    if "class" in df.columns:
        fig = (ggplot(df, aes("model", "score", fill="model"))
                  + geom_col(stat='identity', width=0.6, color="black",
                             position = position_dodge(width=0.2))
                  + ggtitle("Performance Gain compared to Model: " + str(first_model))
                  + facet_grid("metric ~ class")
                  + coord_flip()
                  + xlab("")
                  + ylab("Performance Gain in Percent (%)")
                  + theme_bw()
                  + theme(legend_position="none"))
    # Plot gain results class macro-averaged
    else:
        fig = (ggplot(df, aes("model", "score", fill="model"))
                  + geom_col(stat='identity', width=0.6, color="black",
                             position = position_dodge(width=0.2))
                  + ggtitle("Performance Gain compared to Model: " + str(first_model))
                  + facet_wrap("metric")
                  + coord_flip()
                  + xlab("")
                  + ylab("Performance Gain in Percent (%)")
                  + theme_bw()
                  + theme(legend_position="none"))

    # Store figure to disk
    filename = "plot.comparison.gain"
    if suffix is not None : filename += "." + str(suffix)
    filename += ".png"
    fig.save(filename=filename, path=out_path, width=18, height=9, dpi=300)

    # Return gain dataframe
    return df



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# External Libraries
import numpy as np
import tensorflow as tf
# Internal Libraries
from aucmedi.xai.methods.xai_base import XAImethod_Base
from aucmedi.xai.methods import GuidedBackpropagation, GradCAM
from aucmedi.data_processing.subfunctions import Resize

#-----------------------------------------------------#
#                   Guided Grad-CAM                   #
#-----------------------------------------------------#
class GuidedGradCAM(XAImethod_Base):
    """ XAI Method for Guided Grad-CAM.

    Normally, this class is used internally in the [aucmedi.xai.decoder.xai_decoder][] in the AUCMEDI XAI module.

    ??? abstract "Reference - Implementation"
        Author: Swapnil Ahlawat <br>
        Date: Jul 06, 2020 <br>
        [https://github.com/swapnil-ahlawat/Guided-GradCAM-Keras](https://github.com/swapnil-ahlawat/Guided-GradCAM-Keras) <br>

    ??? abstract "Reference - Publication #1"
        Jost Tobias Springenberg, Alexey Dosovitskiy, Thomas Brox, Martin Riedmiller. 21 Dec 2014.
        Striving for Simplicity: The All Convolutional Net.
        <br>
        [https://arxiv.org/abs/1412.6806](https://arxiv.org/abs/1412.6806)

    ??? abstract "Reference - Publication #2"
        Ramprasaath R. Selvaraju, Michael Cogswell, Abhishek Das, Ramakrishna Vedantam, Devi Parikh, Dhruv Batra. 7 Oct 2016.
        Grad-CAM: Visual Explanations from Deep Networks via Gradient-based Localization.
        <br>
        [https://arxiv.org/abs/1610.02391](https://arxiv.org/abs/1610.02391)

    This class provides functionality for running the compute_heatmap function,
    which computes a Guided Grad-CAM heatmap for an image with a model.
    """
    def __init__(self, model, layerName=None):
        """ Initialization function for creating a Guided Grad-CAM as XAI Method object.

        Args:
            model (keras.model):               Keras model object.
            layerName (str):                   Layer name of the convolutional layer for heatmap computation.
        """
        # Initialize XAI methods
        self.bp = GuidedBackpropagation(model, layerName)
        self.gc = GradCAM(model, layerName)

    #---------------------------------------------#
    #             Heatmap Computation             #
    #---------------------------------------------#
    def compute_heatmap(self, image, class_index, eps=1e-8):
        """ Core function for computing the Guided Grad-CAM heatmap for a provided image and for specific classification outcome.

        ???+ attention
            Be aware that the image has to be provided in batch format.

        Args:
            image (numpy.ndarray):              Image matrix encoded as NumPy Array (provided as one-element batch).
            class_index (int):                  Classification index for which the heatmap should be computed.
            eps (float):                        Epsilon for rounding.

        The returned heatmap is encoded within a range of [0,1]

        ???+ attention
            The shape of the returned heatmap is 2D -> batch and channel axis will be removed.

        Returns:
            heatmap (numpy.ndarray):            Computed Guided Grad-CAM for provided image.
        """
        # Compute Guided Backpropagation
        hm_bp = self.bp.compute_heatmap(image, class_index, eps)
        # Compute Grad-CAM
        hm_gc = self.gc.compute_heatmap(image, class_index, eps)
        hm_gc = Resize(shape=image.shape[1:-1]).transform(hm_gc)
        # Combine both XAI methods
        heatmap = hm_bp * hm_gc
        # Intensity normalization to [0,1]
        numer = heatmap - np.min(heatmap)
        denom = (heatmap.max() - heatmap.min()) + eps
        heatmap = numer / denom
        # Return the resulting heatmap
        return heatmap



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# External libraries
from sklearn.utils.class_weight import compute_class_weight, compute_sample_weight
import numpy as np

#-----------------------------------------------------#
#               Class Weight Computation              #
#-----------------------------------------------------#
def compute_class_weights(ohe_array, method="balanced"):
    """ Simple wrapper function for scikit learn class_weight function.

    The class weights can be used for weighting the loss function on imbalanced data.

    ???+ info
        NumPy array shape has to be (n_samples, n_classes) like this: (500, 4).

    ???+ example "Example"
        ```python
        # Compute class weights
        cw_loss, cw_fit = compute_class_weights(class_ohe)

        # Provide class weights to loss function
        model = NeuralNetwork(nclasses, channels=3, loss=categorical_focal_loss(cw_loss))

        # Provide class weights to keras fit()
        model.train(index_list, epochs=50, class_weights=cw_fit)
        ```

    ??? abstract "Based on Reference"
        Scikit-learn class_weight function: <br>
        https://scikit-learn.org/stable/modules/generated/sklearn.utils.class_weight.compute_class_weight.html  <br>

    Args:
        ohe_array (numpy.ndarray):  NumPy matrix containing the ohe encoded classification.
        method (str):               Dictionary or modus, how class weights should be computed.

    Returns:
        class_weights_list (numpy.ndarray):     Class weight list which can be feeded to a loss function.
        class_weights_dict (dict):              Class weight dictionary which can be feeded to
                                                [train()][aucmedi.neural_network.model.NeuralNetwork.train]
                                                or keras.model.fit().
    """
    # Obtain sparse categorical array and number of classes
    class_array = np.argmax(ohe_array, axis=-1)
    n_classes = np.unique(class_array)
    # Compute class weights with scikit learn
    class_weights_list = compute_class_weight(class_weight=method,
                                              classes=n_classes, y=class_array)
    # Convert class weight array to dictionary
    class_weights_dict = dict(enumerate(class_weights_list))
    # Return resulting class weights as list and dictionary
    return class_weights_list, class_weights_dict

#-----------------------------------------------------#
#           Multi-Label Weight Computation            #
#-----------------------------------------------------#
def compute_multilabel_weights(ohe_array, method="balanced"):
    """ Function for computing class weights individually for multi-label data.

    Class weights can be used for weighting the loss function on imbalanced data.
    Returned is a class weight list which can be passed to loss functions.

    ???+ info
        NumPy array shape has to be (n_samples, n_classes) like this: (500, 4).

    ??? abstract "Based on Reference"
        Class weight computation is based on Scikit learn class_weight function: <br>
        https://scikit-learn.org/stable/modules/generated/sklearn.utils.class_weight.compute_class_weight.html  <br>

    Args:
        ohe_array (numpy.ndarray):      NumPy matrix containing the ohe encoded classification.
        method (str):                   Dictionary or modus, how class weights should be computed.

    Returns:
        class_weights (numpy.ndarray):      Class weight list which can be fed to a loss function.
    """
    # Identify number of classes
    n_classes = np.shape(ohe_array)[1]
    # Initialize class weight list
    class_weights = np.empty([n_classes])
    # Compute weight for each class individually
    for i in range(0, n_classes):
        weight = compute_class_weight(class_weight=method, classes=np.array([0,1]),
                                      y=ohe_array[:, i])
        class_weights[i] = weight[1]
    # Return resulting class weight list
    return class_weights

#-----------------------------------------------------#
#              Sample Weight Computation              #
#-----------------------------------------------------#
def compute_sample_weights(ohe_array, method="balanced"):
    """ Simple wrapper function for scikit learn sample_weight function.

    The sample weights can be used for weighting the loss function on imbalanced data.
    Returned sample weight array which can be directly fed to an AUCMEDI [DataGenerator][aucmedi.data_processing.data_generator.DataGenerator].

    ???+ info
        NumPy array shape has to be (n_samples, n_classes) like this: (500, 4).

    ??? abstract "Based on Reference"
        Scikit learn sample_weight function: <br>
        https://scikit-learn.org/stable/modules/generated/sklearn.utils.class_weight.compute_sample_weight.html <br>

    Args:
        ohe_array (numpy.ndarray):      NumPy matrix containing the ohe encoded classification.
        method (str):                   Dictionary or modus, how class weights should be computed.

    Returns:
        sample_weights (numpy.ndarray):     Sample weight list which can be fed to an AUCMEDI
                                            [DataGenerator][aucmedi.data_processing.data_generator.DataGenerator].
    """
    # Compute sample weights with scikit learn
    sample_weights = compute_sample_weight(class_weight=method, y=ohe_array)
    # Return resulting sample weights
    return sample_weights
